/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/*	@(#)AR_cmd.c	1.2 9/18/92
 *	Author:	Roger A. Cole
 *	Date:	07-28-92
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 * make options
 *	-DNDEBUG	don't compile assert() checking
 *	-DDEBUG		compile various debug code, including checks on
 *			malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	AR_cmd - EPICS archiver with keyboard interface
*
* DESCRIPTION
*
*
* BUGS
* o	much of the code here duplicates the functionality of corresponding
*	code in AR.c, but with enough differences to disallow sharing
* o	not all signals are caught
*
*-***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <ctype.h>
#include <signal.h>
#include <sys/fcntl.h>

#include <genDefs.h>
#include <envDefs.h>
#include <cmdDefs.h>
#include <cadef.h>

#include <arCS.h>
#include <arAccessLib.h>
#include <sydDefs.h>
#include <reqSubr.h>

/*-----------------------------------------------------------------------------
*
*----------------------------------------------------------------------------*/
typedef enum {
    REQ_ST_CONN=1,
    REQ_ST_NOT_CONN,
    REQ_ST_ENABLED,
    REQ_ST_ACTIVE,
    REQ_ST_INACTIVE,
} REQ_ST;

typedef enum {
    AR_INF_STATE,		/* show request state */
    AR_INF_CONN,		/* show connection information */
    AR_INF_SAMP,		/* show sampling information */
    AR_INF_STEP,		/* show `step a channel' information */
} AR_INF;


typedef struct arChan {
    struct arChan *pNext;	/* pointer to next channel */
    struct arChan *pPrev;	/* pointer to previous channel */
    AR_CHAN_DESC *pCfChan;	/* pointer to `bcf' channel descriptor */
    SYD_CHAN	*pSChan;	/* pointer to syd channel descriptor */
    struct arReq *pReq;		/* pointer to archiver request for channel */
    short	getGR;		/* 1 if need to get graphics info */
} AR_CHAN;

typedef struct arReq {
    struct arReq *pNext;
    struct arReq *pPrev;
    struct arCtx *pCtx;
    char	name[db_name_dim];/* name of archive request */
    char	file[120];	/* file name for request */
    char	dir[120];	/* directory for request */
    REQ_ST	status;		/* status of the request */
    short	nDis;		/* number of channels disconnected */
    short	nCtrlDis;	/* number of control channels disconnected */
    short	sampling;	/* 0 if no sampling in progress;
				   1 if actual sampling is in progress;
				   -1 if pre-trigger queueing is in progress */
    SYD_SPEC	*pSspec;
    REQ_TRIG	actTrig;	/* activation trigger */
    short	allowDiscon;	/* allow disconnected channels */
    REQ_TRIG	inactTrig;	/* inactivation trigger */
    REQ_SAMPLE	sample;		/* sample method structure */
    AR_CF_DESC	*pCfDesc;	/* by channel file handle */
    AR_CHAN	*pChanHead;	/* head of channel list */
    AR_CHAN	*pChanTail;	/* tail of channel list */
    char	lockPath[250];	/* lock on open */
    short	getGR;		/* 1 if need to get graphics info */
    TS_STAMP	firstTs;	/* time of first data, while monitoring */
    TS_STAMP	lastTs;		/* time of last data, while monitoring */
    int		monitorCount;	/* count of values, while monitoring */
    REQ_STEP	step;		/* `step a channel' structure */
} AR_REQ;

typedef struct arCtx {
    CX_CMD	cmdCtx;
    CX_CMD	*pCmdCtx;
    AR_REQ	*pReqHead;	/* list of opened archive requests */
    AR_REQ	*pReqTail;
    jmp_buf     sigEnv;         /* environment for longjmp at signal time */
    int		IPPort;		/* InterNet port number for server */
    int		serverMode;	/* 1 says running as server */
    int         listenSock;     /* fd for server's "listen" socket */
    int         inUse;          /* 1 says "presently engaged" by a client */
    int         inUseInit;      /* 1 says in-use init has been done */
    char        *inUseName;     /* name of client's host */
    int         clientSock;     /* fd for socket to client */
    int         reopenIO;       /* std... should be re-directed */
    int         reopenIOoption; /* 0,1,2; option for arFreopenAllStd */
    int         oldStdinFd;     /* original fd for stdin, before redirect */
    int         oldStdoutFd;    /* original fd for stdout, before redirect */
    int         oldStderrFd;    /* original fd for stderr, before redirect */
    char        *oldPrompt;	/* prompt prior to socket connect */
    int		stop;		/* 1 says stop operations */
} AR_CTX;

/*/subhead -------------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
void arCmdCrunch();
void arListenSetReopen();
long arInitAtStartup();
void arSigHandler();
AR_REQ *arReqFind();
void arReqSampleMonHandler();
void arTimer();
long arListenClose();
long arListenQuit();

/*-----------------------------------------------------------------------------
* global definitions
*----------------------------------------------------------------------------*/
static char	*progName = "AR_cmd";
static AR_CTX	*pglArCtx;

#define AR_PRINTF arLogPrintf(pCtx,

/* forward definition */
#ifdef __STDC__
void arLogPrintf(AR_CTX *pCtx, char *fmt, ...);
#else
void arLogPrintf();
#endif


/*+/subr**********************************************************************
* NAME	main - AR_kbd main routine
*
* DESCRIPTION
*       % AR_cmd              (runs as normal interactive program)
*       % AR_cmd server       (runs as a server)
*       % AR_cmd server&      (runs as a server)
*       % AR_cmd hostName     (runs as a client to AR_cmd on hostName)
*
* BUGS
* o	need a place to send output when running detached.  Maybe to AR log.
* o	need a graceful way to wind down operations at exit
* o	doesn't optimize channel access gets and monitors--if several
*	requests get data from the same channel, then several gets or
*	monitors will be done
*
*-*/
main(argc, argv)
int     argc;		/* number of command line args */
char   *argv[];		/* command line args */
{
    long	stat;		/* status return from calls */
    int		sigNum;
    int		i;
    AR_REQ	*pReq;
    int		syncCount;
    char	portText[10];
    AR_CTX	arCtx, *pCtx=&arCtx;

    stat = arInitAtStartup(pCtx);
    assert(stat == OK);
    pglArCtx = pCtx;

    if (argc > 2) {
	printf("Usage: %s [server or hostName]\n", argv[0]);
	exit(1);
    }
    else if (argc == 2) {
	if (envGetConfigParam(&EPICS_AR_PORT, 10, portText) == NULL) {
	    printf("error getting %s\n", EPICS_AR_PORT.name);
	    return ERROR;
	}
	sscanf(portText, "%d", &pCtx->IPPort);
	if (strcmp(argv[1], "server") != 0) {
	    execlp("cmdClient", "cmdClient", argv[1], portText, (char *)0);
	    (void)printf("couldn't exec to cmdClient\n");
	    exit(1);
	}
	pCtx->serverMode = 1;
	pCtx->inUseName = argv[1];
/*-----------------------------------------------------------------------------
*       establish a socket to use to listen for potential clients to connect
*----------------------------------------------------------------------------*/
	if (ezsCreateListenSocket(&pCtx->listenSock, pCtx->IPPort) != 0) {
	    perror("create listen socket");
	    exit(1);
	}
    }
    else
	pCtx->inUseName = "shell";

/*-----------------------------------------------------------------------------
* we're going to be a "real" archiver--either interactive or a server.  Make
* sure there will be only one of us in this directory.
*----------------------------------------------------------------------------*/
    if (arLock(pCtx, "archiver running here", "AR.lock", 1) != 1) {
	(void)printf(
"\nAR.lock indicates that another copy of AR is running or else the lock\n\
file has inadvertently been left (perhaps by an abort).  You will need\n\
to diagnose which.\n\n");
	exit(1);
    }

    AR_PRINTF "starting");

    genSigInit(arSigHandler);
    if ((sigNum = setjmp(pCtx->sigEnv)) != 0) {
	AR_PRINTF "signal received: %d", sigNum);
	goto arTaskWrapup;
    }

/*-----------------------------------------------------------------------------
*    in server mode read all opened requests at startup time
*----------------------------------------------------------------------------*/
    if (pCtx->serverMode) {
	stat = arReadEnabled(pCtx);
	if (stat != OK)
	    exit(1);
    }

/*----------------------------------------------------------------------------
*    "processing loop"
*---------------------------------------------------------------------------*/
    syncCount = 0;
    while (!pCtx->stop) {
	if (pCtx->serverMode)
	    arListenCheck(pCtx);
	if (!pCtx->serverMode || pCtx->inUse) {
	    if (cmdRead(&pCtx->pCmdCtx) != NULL) {
		if (pCtx->serverMode && pCtx->pCmdCtx->inputEOF) {
		    pCtx->pCmdCtx->pCommand = "disconnect";
		    pCtx->pCmdCtx->inputEOF = 0;
		}
		arCmdCrunch(pCtx);
	    }
	}
	arTimer(pCtx);
	if (++syncCount >= 100) {
	    syncCount = 0;
	    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
		if (pReq->pCfDesc != NULL)
		    stat = arCFSyncWrite(pReq->pCfDesc);
	    }
	}
        fflush(stdout);         	/* flushes needed for socket I/O */
        fflush(stderr);
/*  usleep isn't POSIX */
#if defined(SOLARIS) || defined(HP_UX) || defined(SGI)
      sleep(1);
#else
	usleep(100000);		/* sleep .1 second */
#endif
    }

arTaskWrapup:
    pCtx->stop = 1;
    if (pCtx->reopenIO) {
	(void)arFreopenAllStd(pCtx, pCtx->reopenIOoption);
	pCtx->reopenIO = 0;
    }
    arListenQuit(pCtx);

    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext)
	stat = arReqClose(pReq);

    AR_PRINTF "stopping");

    fflush(stdout);         /* flushes needed for sockets */
    fflush(stderr);

    arLock(pCtx, "archiver running here", "AR.lock", 0);

    exit(0);
}

/*+/subr**********************************************************************
* NAME	arTimer - handle notifier timer calls
*
*-*/
void
arTimer(pCtx)
AR_CTX	*pCtx;
{
    AR_REQ	*pReq;
    SYD_CHAN	*pChan;
    int		nCon, nDis, nCtrlDis;
    TS_STAMP	nowTs;
    double	diff;
    static TS_STAMP last1secTs={0,0}, last15secTs={0,0};
    long	stat;

    ca_pend_event(.001);
    tsLocalTime(&nowTs);
    TsDiffAsDouble(&diff, &nowTs, &last1secTs);
    if (diff >= 1.) {
/*-----------------------------------------------------------------------------
*    1 second processes
*    o  check channel connections, and update status on changes
*    o  check to activate request
*    o  check to inactivate request
*----------------------------------------------------------------------------*/
	last1secTs = nowTs;
	for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
	    arReqCheckConn(pReq, &nCon, &nDis, &nCtrlDis, 0);
	    if (nDis != pReq->nDis) {
		if (nDis > 0) {
		    AR_PRINTF "%s: connects pending: total=%d, control=%d",
				pReq->name, nDis, nCtrlDis);
		}
		else
		    AR_PRINTF "%s: connections are complete", pReq->name);
		if (nDis == 0 || (nCtrlDis == 0 && pReq->allowDiscon)) {
		    pReq->status = REQ_ST_CONN;
		    reqTrigArm(&pReq->actTrig, nowTs);
		}
		pReq->nDis = nDis;
		pReq->nCtrlDis = nCtrlDis;
	    }
	    arReqCheckAct(pReq, nowTs);
	    arReqCheckInact(pReq, nowTs);
	}
	TsDiffAsDouble(&diff, &nowTs, &last15secTs);
	if (diff >= 15.) {
/*-----------------------------------------------------------------------------
*    15 second processes
*    o  if awaiting graphics information for `by channel' output, check for
*	its arrival and store it when available
*    o  sync `by channel' files
*----------------------------------------------------------------------------*/
	    last15secTs = nowTs;
	    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
		if (pReq->getGR)
		    arReqCheckGraphics(pReq);
		if (pReq->sample.outputType == REQ_OUT_BCF) {
		    stat = arCFSyncWrite(pReq->pCfDesc);
		    assert(stat == OK);
		}
	    }
	}
    }
/*-----------------------------------------------------------------------------
*    .1 second processes
*    o  ca_pend_event		(called above)
*    o  check for sampling activities for request (if the request has
*	inactivation type of REQ_TYPE_DONE and sampling is complete, then
*	this check is skipped)
*    o  check for `step a channel' activities.  Most of the work is done
*	here, but SnapStart and MonStart help with detecting when sampling
*	actually starts (for TRIGGER: after each step).
*----------------------------------------------------------------------------*/
    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
	if (pReq->status == REQ_ST_ACTIVE && (pReq->inactTrig.doneFlag == 0 ||
				pReq->inactTrig.type != REQ_TYPE_DONE)) {
	    arReqCheckSample(pReq, nowTs);
	}
	if (pReq->step.active) {
	    if (pReq->step.sampleActive) {
		if (!pReq->sampling)
		    pReq->step.sampleActive = 0, pReq->step.sampleDone = 1;
	    }
	    if (reqStep_stepDoneCheck(&pReq->step, &nowTs)) {
		if (pReq->sample.trigger.type == REQ_TYPE_AFTER_STEP &&
						pReq->step.sampleDone == 0) {
		    pReq->sample.trigger.demand = 1;
		    pReq->step.samplePending = 1;
		    if (pReq->status == REQ_ST_ACTIVE &&
				(pReq->inactTrig.doneFlag == 0 ||
				pReq->inactTrig.type != REQ_TYPE_DONE)) {
			arReqCheckSample(pReq, nowTs);
		    }
		}
		else if (!pReq->step.complete) {
		    reqStep_stepNext(&pReq->step);
		    pReq->step.sampleDone = 0;
		}
		reqStep_seqDoneCheck(&pReq->step);
	    }
	}
    }
}

/*+/subr**********************************************************************
* NAME	arReqCheckAct - check for need to activate a request
*
* DESCRIPTION
*	Checks to see if it's OK to activate a request.  If it is OK, and
*	if the activation condition is met, the request is activated.
*
*	There are several special cases which keep a request from activating:
*	o   the request is waiting to receive graphics information to be
*	    stored in a `by channel' archive file
*	o   all the channels in the request aren't connected; this includes
*	    trigger channels
*	o   an event trigger for a DBF_ENUM channel is waiting to receive
*	    graphics information (to get the state strings)
*	o   an event trigger for a DBF_ENUM channel has a comparison value
*	    which isn't one of the state strings for the channel
*
* BUGS
* o	doesn't put an "activated" item in `by channel' file
*
*-*/
arReqCheckAct(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    AR_CHAN	*pRChan;

    if (pReq->getGR) {
	arReqCheckGraphics(pReq);
	if (pReq->getGR)
	    return;
    }
    if (pReq->status == REQ_ST_CONN || pReq->status == REQ_ST_INACTIVE) {
	if (pReq->actTrig.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->actTrig) != 0)
		return;
	}
	if (pReq->inactTrig.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->inactTrig) != 0)
		return;
	}
	if (pReq->sample.forEvent.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->sample.forEvent) != 0)
		return;
	}
	if (pReq->sample.trigger.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->sample.trigger) != 0)
		return;
	}
	if (reqTrigTest(&pReq->actTrig, nowTs)) {
	    pReq->status = REQ_ST_ACTIVE;
	    AR_PRINTF "%s: activated", pReq->name);
	    reqTrigArm(&pReq->inactTrig, nowTs);
	}
    }
}

/*+/subr**********************************************************************
* NAME	arReqCheckConn - check connections for a request's channels
*
*-*/
arReqCheckConn(pReq, pNcon, pNdiscon, pNCtrlDiscon, printFlag)
AR_REQ	*pReq;
int	*pNcon, *pNdiscon, *pNCtrlDiscon, printFlag;
{
    SYD_CHAN	*pChan;

    *pNcon=0;
    *pNdiscon=0;
    *pNCtrlDiscon=0;
    for (pChan=pReq->pSspec->pChanHead; pChan!= NULL; pChan=pChan->pNext){
	if (pChan->discon) {
	    (*pNdiscon)++;
	    if (printFlag)
		printf("      %s\n", pChan->name);
	}
	else
	    (*pNcon)++;
    }
    if (pReq->actTrig.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->actTrig.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->actTrig.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->inactTrig.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->inactTrig.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->inactTrig.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->sample.forEvent.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->sample.forEvent.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->sample.trigger.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->sample.trigger.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->sample.trigger.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->step.type == REQ_TYPE_STEP) {
	if (ca_field_type(pReq->step.pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->step.pCh));
	}
	else
	    (*pNcon)++;
    }
}

/*+/internal******************************************************************
* NAME	arReqCheckEnum - check enumerated trigger channel
*
* DESCRIPTION
*	Checks to see if graphics info for enumerated channels has been
*	received.  If so, checks to see that the comparison value is legal.
*
* RETURNS
*	0	if comparison value is OK, or
*	-1	if graphics info hasn't yet been received, or
*	-2	if comparison value is illegal
*
* BUGS
* o	failure should set an invalid state or close the request;
*	presently, the only action is to give a notice and let the
*	operator do "whatever", including nothing.
*
*-*/
arReqCheckEnum(pReq, pTrig)
AR_REQ	*pReq;
REQ_TRIG *pTrig;
{
    AR_CTX	*pCtx=pReq->pCtx;
    int		i;

    assert(pTrig->enumGetGR != 0);
    if (pTrig->enumGetGR != 2)
	return -1;
    pTrig->enumGetGR = 0;
    for (i=0; i<pTrig->enumGr.no_str; i++) {
	if (strcmp(pTrig->eventValCmp.strVal, pTrig->enumGr.strs[i]) == 0)
	    return 0;
    }
    AR_PRINTF "condition test is invalid; \"%s\" not a legal value for \"%s\"",
pTrig->eventValCmp.strVal, pTrig->eventName);
    return -2;
}

/*+/internal******************************************************************
* NAME	arReqCheckGraphics
*
* DESCRIPTION
*    if graphics information was requested for a channel being written to
*    a `by channel' file, and if the graphics information has arrived,
*    it is written to the file.
*
*-*/
arReqCheckGraphics(pReq)
AR_REQ	*pReq;
{
    AR_CHAN	*pRChan;
    struct event_handler_args arg;
    long	stat;

    for (pRChan=pReq->pChanHead; pRChan!=NULL; pRChan=pRChan->pNext) {
	if (pRChan->getGR && pRChan->pSChan->gotGr) {
	    arg.dbr = &pRChan->pSChan->grBuf;
	    arg.count = pRChan->pSChan->elCount;
	    arg.chid = (chid)pRChan->pSChan->pHandle;
	    arg.type = pRChan->pSChan->dbrGrType;
	    stat = arCFChanWriteGR(pRChan->pCfChan, &arg);
	    assert(stat == OK);
	    pRChan->getGR = 0;
	    pReq->getGR--;
	}
    }
}

/*+/subr**********************************************************************
* NAME	arReqCheckInact - check for need to inactivate a request
*
* DESCRIPTION
*	o   check for inactivation condition
*
*-*/
arReqCheckInact(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    int		doIt=0;

    if (pReq->status != REQ_ST_ACTIVE)
	return;
    if (pReq->nCtrlDis != 0 || (pReq->allowDiscon == 0 && pReq->nDis != 0)) {
	AR_PRINTF "%s: inactivated; %d broken connections",
						pReq->name, pReq->nDis);
	doIt = 1;
	pReq->status = REQ_ST_NOT_CONN;
    }
    else {
	if (reqTrigTest(&pReq->inactTrig, nowTs)) {
	    AR_PRINTF "%s: inactivated", pReq->name);
	    doIt = 1;
	    pReq->status = REQ_ST_INACTIVE;
	    reqTrigArm(&pReq->actTrig, nowTs);
	}
    }
    if (doIt == 0)
	return;
    if (pReq->sample.type == REQ_TYPE_SNAP)
	arReqSampleSnapStop(pReq, nowTs);
    else
	arReqSampleMonStop(pReq, nowTs);
}

/*+/subr**********************************************************************
* NAME	arReqCheckSample - check sampling activities
*
* DESCRIPTION
*	o   check for sample trigger
*	o   check for sample condition
*	o   check for "until" condition
*	o   check for need to handle "before trigger" condition
*
* BUGS
* o	for "n samples before trigger", actual sample can start as much
*	as 10 samples prior to desired spot.  This is due to the fact that
*	the trigger comes directly from a CA monitor, and the sync samples
*	are buffered some time later.  This could be improved some by
*	moving the test to follow the sydInputGet loop.  An exact solution
*	requires that the trigger channel be part of the sync set and the
*	test based on the values in the syd buffers.
* o	for "until condition", sampling stops as soon as the condition is
*	satisfied.  This may leave several samples in the syd buffers if
*	filled and sync data are being intermixed.
*
*-*/
arReqCheckSample(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;

/*-----------------------------------------------------------------------------
*    Check to see if any sampling activities need to be started.
*
*	If ordinary sampling is being done, then no activities are needed
*	until the trigger occurs.
*
*	If pre-trigger sampling is being done, then the pre-trigger queueing
*	needs to be started.
*----------------------------------------------------------------------------*/
    if (pReq->sampling == 0) {
	if (pReq->sample.trigger.armed != 1) {
	    reqTrigArm(&pReq->sample.trigger, nowTs);
	}
	if (reqTrigTest(&pReq->sample.trigger, nowTs) == 0) {
	    if (pReq->sample.startWhen != REQ_TYPE_BEFORE)
		return;
	    if (pReq->sample.trigger.type != REQ_TYPE_EVENT ||
				pReq->sample.trigger.everyPrimed == 1) {
		arReqSampleSnapStart(pReq, nowTs);
	    }
	    else
		return;
	}
	else if (pReq->sample.type == REQ_TYPE_SNAP) {
	    arReqSampleSnapStart(pReq, nowTs);
	    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT)
		reqTrigArm(&pReq->sample.forEvent, nowTs);
	}
	else {
	    arReqSampleMonStart(pReq, nowTs);
	    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT)
		reqTrigArm(&pReq->sample.forEvent, nowTs);
	}
    }
    else if (pReq->sampling == -1) {
	if (reqTrigTest(&pReq->sample.trigger, nowTs) == 1) {
/*-----------------------------------------------------------------------------
*	A trigger has finally occurred in pre-trigger queueing mode.  Set
*	up the proper starting sample and switch into normal sampling mode.
*
*	First, figure out which sample to start with, depending on whether
*	pre-trigger is based on number of samples or on time.  Then, adjust
*	the deltaSec in the sample set spec, and also adjust firstData and
*	sampleCount.  If sampling is "for hh:mm:ss", adjust forStamp.
*----------------------------------------------------------------------------*/
	    TS_STAMP	ts;
	    short	samp=pSspec->firstData;

	    if (pSspec->sampleCount <= 0)
		;	/* no samples yet; just switch into normal mode */
	    else if (pReq->sample.startType == REQ_TYPE_NTH) {
		if (pSspec->sampleCount <= pReq->sample.startCount)
		    ;	/* current count <= desired; just go to normal mode */
		else {
		    samp = pSspec->lastData - pReq->sample.startCount + 1;
		    if (samp < 0)
			samp += pSspec->dataDim;
		}
	    }
	    else {
		TsDiffAsStamp(&ts,
			&pReq->sample.trigger.dataTs, &pReq->sample.startStamp);
		while (1) {
		    if (TsCmpStampsGE(&pSspec->pTimeStamp[samp], &ts))
			break;
		    if (samp == pSspec->lastData)
			break;
		    else if (++samp >= pSspec->dataDim)
			samp = 0;
		}
	    }
	    if (samp != pSspec->firstData) {
		ts = pSspec->pTimeStamp[samp];
		ts.nsec = 0;
		pSspec->refTs = ts;
		pSspec->firstData = samp;
		pSspec->sampleCount = 0;
		while (1) {
		    pSspec->sampleCount++;
		    TsDiffAsDouble(&pSspec->pDeltaSec[samp],
						&pSspec->pTimeStamp[samp], &ts);
		    if (samp == pSspec->lastData)
			break;
		    else if (++samp >= pSspec->dataDim)
			samp = 0;
		}
	    }
	    if (pReq->sample.forType == REQ_TYPE_TIMED) {
		double	d=-1.*pSspec->pDeltaSec[pSspec->lastData];
		TsAddStamps(&pReq->sample.forStop,
						&nowTs, &pReq->sample.forStamp);
		TsAddDouble(&pReq->sample.forStop, &pReq->sample.forStop, d);
	    }
	    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT)
		reqTrigArm(&pReq->sample.forEvent, nowTs);
	    pReq->sampling = 1;
	    AR_PRINTF "%s: snapshot sampling...", pReq->name);
	}
    }
    if (pReq->sample.halt == 1)
	goto checkSample_stop;
    if (pReq->sample.type == REQ_TYPE_SNAP) {
	while (1) {
	    if ((stat = sydInputGet(pSspec, NULL)) == S_syd_noDataNow)
		break;
	    sydInputStoreInSet(pSspec, pReq->sample.ignorePartial);
	    if (pReq->sampling==1 && pReq->sample.forType==REQ_TYPE_SAMPLES) {
		if (pSspec->sampleCount >= pReq->sample.forCmp)
		    goto checkSample_stop;
	    }
	    if (pSspec->sampleCount >= pSspec->reqCount)
		goto checkSample_stop;
	}
    }
    else if (pReq->sample.type == REQ_TYPE_MONITOR) {
	if (pReq->sample.forType == REQ_TYPE_SCHED) {
	    if (TsCmpStampsGT(&pReq->lastTs, &pReq->sample.forStop))
		goto checkSample_stop;
	}
    }
    if (pReq->sampling == 1) {
/*-----------------------------------------------------------------------------
*    Check to see if sampling should be stopped.  This might be for any of:
*	o   for n samples  (only for SNAP, checked above)
*	o   for hh:mm:ss
*	o   until chanName condition
*
*	These checks aren't made if pre-trigger queueing is going on.
*----------------------------------------------------------------------------*/
	if (pReq->sample.forType == REQ_TYPE_TIMED &&
			TsCmpStampsGT(&nowTs, &pReq->sample.forStop)) {
	    goto checkSample_stop;
	}
	if (pReq->sample.forType == REQ_TYPE_EVENT) {
	    if (reqEventTest_sm(&pReq->sample.forEvent, NULL) == 1) {
		goto checkSample_stop;
	    }
	}
    }
    return;

checkSample_stop:
    if (pReq->sample.type == REQ_TYPE_SNAP)
	arReqSampleSnapStop(pReq, nowTs);
    else
	arReqSampleMonStop(pReq, nowTs);
}

/*+/subr**********************************************************************
* NAME	arSigHandler - signal handling
*
* DESCRIPTION
*       These routines set up for the signals to be caught for arTask
*       and handle the signals when they occur.
*
* RETURNS
*       void
*
* BUGS
* o     not all signals are caught
* o     it's not clear how useful it is to catch the signals which originate
*       from the keyboard
*
*-*/
static void
arSigHandler(signo)
int     signo;
{
    if (signo == SIGPIPE) {
#if 0
	if (arFreopenAllStd(pglArCtx, 0) != OK)
	    assertAlways(0);
	if (!pglArCtx->stop) {
	    (void)printf("client connection broken\n");
	    pglArCtx->arInTaskInfo.stop = 1;
	}
#endif
	return;
    }
    printf("entered arSigHandler for signal:%d\n", signo);
    signal(signo, SIG_DFL);
    longjmp(pglArCtx->sigEnv, signo);
}

/*+/subr**********************************************************************
* NAME	arCmdCrunch - process a command line
*
* DESCRIPTION
*
* RETURNS
*	void
*
* BUGS
* o	should allow "help command" to get more information
*
*-*/
void
arCmdCrunch(pCtx)
AR_CTX *pCtx;	/* IO pointer to ar context */
{
    CX_CMD	*pCmdCtx=pCtx->pCmdCtx;/* local copy of pointer */
    int		stat;		/* status return */

    chid	pCh;		/* channel pointer */
    AR_REQ	*pReq;		/* request pointer */
    AR_REQ	*arReqFind();	/* find named request on list */
    char	reqName[AR_NAME_DIM];/* request name */

#define IF_NEXT_NON_SPACE_FIELD_IS_NULL if (nextNonSpaceField( \
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1)

    if (strcmp(pCmdCtx->pCommand,			"assert0") == 0) {
/*-----------------------------------------------------------------------------
*	under SunOS, this generates SIGABRT
*----------------------------------------------------------------------------*/
	assertAlways(0);
    }
    else if (strcmp(pCmdCtx->pCommand,			"quit") == 0) {
	pCtx->stop = 1;
	if (pCtx->serverMode)
	    arListenClose(pCtx);
    }
    else if (strcmp(pCmdCtx->pCommand,			"trap") == 0) {
	int	*j;
	j = (int *)(-1);
	j = (int *)(*j);
    }
    else if (strcmp(pCmdCtx->pCommand,			"disconnect") == 0) {
	if (strcmp(pCtx->inUseName, "shell") == 0)
	    (void)printf("disconnect illegal--AR not running as server\n");
	else if (pCtx->serverMode) {
	    arListenClose(pCtx);
	}
    }
    else if (strcmp(pCmdCtx->pCommand,			"activate") == 0) {
/*----------------------------------------------------------------------------
* activate - activate a particular archive request
*---------------------------------------------------------------------------*/
	IF_NEXT_NON_SPACE_FIELD_IS_NULL {
	    (void)printf("usage: activate reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCtx, pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->actTrig.type != REQ_TYPE_DEMAND) {
	    (void)printf("activation type not `demand'\n");
	    return;
	}
	pReq->actTrig.demand = 1;
    }
    else if (strcmp(pCmdCtx->pCommand,			"close") == 0) {
/*----------------------------------------------------------------------------
* close - close a particular archive request or all archive requests
*---------------------------------------------------------------------------*/
	IF_NEXT_NON_SPACE_FIELD_IS_NULL
	    (void)printf("usage: close reqName  or  all\n");
	else if (strcmp(pCmdCtx->pField, "all") == 0) {
	    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
		stat = arReqClose(pReq);
	    }
	}
	else {
	    if ((pReq = arReqFind(pCtx, pCmdCtx->pField)) == NULL) {
		(void)printf("couldn't find %s\n", pCmdCtx->pField);
		return;
	    }
	    stat = arReqClose(pReq);
	}
    }
    else if (strcmp(pCmdCtx->pCommand,			"open") == 0) {
/*----------------------------------------------------------------------------
* open - read an archive request from file in ./arReq/name.req or all
*	the requests in ./arReq/open.dat
*---------------------------------------------------------------------------*/
	IF_NEXT_NON_SPACE_FIELD_IS_NULL
	    (void)printf("usage: open reqName  or  all\n");
	else if (strcmp(pCmdCtx->pField, "all") == 0)
	    stat = arReadEnabled(pCtx);
	else
	    stat = arReqOpen(pCtx, pCmdCtx->pField);
    }
    else if (strcmp(pCmdCtx->pCommand,			"inactivate") == 0) {
/*----------------------------------------------------------------------------
* inactivate - inactivate a particular archive request
*---------------------------------------------------------------------------*/
	IF_NEXT_NON_SPACE_FIELD_IS_NULL {
	    (void)printf("usage: inactivate reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCtx, pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->inactTrig.type != REQ_TYPE_DEMAND) {
	    (void)printf("inactivation type not `demand'\n");
	    return;
	}
	if (pReq->status != REQ_ST_ACTIVE) {
	    (void)printf("request isn't active\n");
	    return;
	}
	pReq->inactTrig.demand = 1;
    }
    else if (strcmp(pCmdCtx->pCommand,			"sample") == 0) {
/*----------------------------------------------------------------------------
* sample - sample a particular archive request
*---------------------------------------------------------------------------*/
	IF_NEXT_NON_SPACE_FIELD_IS_NULL {
	    (void)printf("usage: sample reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCtx, pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->sample.trigger.type != REQ_TYPE_DEMAND) {
	    (void)printf("sample type not `demand'\n");
	    return;
	}
	if (pReq->status != REQ_ST_ACTIVE) {
	    (void)printf("request isn't active\n");
	    return;
	}
	pReq->sample.trigger.demand = 1;
    }
    else if (strcmp(pCmdCtx->pCommand,			"startStepping") == 0) {
/*----------------------------------------------------------------------------
* startStepping - start a sequence of steps for a particular archive request
*---------------------------------------------------------------------------*/
	IF_NEXT_NON_SPACE_FIELD_IS_NULL {
	    (void)printf("usage: startStepping reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCtx, pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->step.active) {
	    (void)printf("already stepping %s\n", pCmdCtx->pField);
	    return;
	}
	if (ca_field_type(pReq->step.pCh) == TYPENOTCONN) {
	    (void)printf("channel disconnected: %s\n", pReq->step.name);
	    return;
	}
	reqStep_seqStart(&pReq->step);
    }
    else if (strcmp(pCmdCtx->pCommand,			"status") == 0) {
/*----------------------------------------------------------------------------
* status - print archiver status
*---------------------------------------------------------------------------*/
	if (pCtx->pReqHead == NULL)
	    AR_PRINTF "no opened requests");
	else {
	    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
		AR_PRINTF "%s", pReq->name);
		arReqStatus(pReq, AR_INF_STATE);
		arReqStatus(pReq, AR_INF_CONN);
		arReqStatus(pReq, AR_INF_SAMP);
		arReqStatus(pReq, AR_INF_STEP);
	    }
	}
    }
    else {
/*----------------------------------------------------------------------------
* help (or illegal command)
*----------------------------------------------------------------------------*/
	(void)nextNonSpaceField(&pCmdCtx->pLine, &pCmdCtx->pField,
							&pCmdCtx->delim);
	helpIllegalCommand(stdout, &pCmdCtx->helpList, pCmdCtx->pCommand,
							pCmdCtx->pField);
    }
}

/*+/subr**********************************************************************
* NAME	arListenCheck - listen for client connections
*
* DESCRIPTION
*
* RETURNS
*
* BUGS
* o     cleanup of listen socket isn't dealt with.  Under SunOS, the system
*       seems to clean up OK.
*
*       In fact, under SunOS, shutdown and/or close seem to inhibit proper
*       cleanup by the system.
*
* SEE ALSO
*
* EXAMPLE
*
*-*/
int
arListenCheck(pCtx)
AR_CTX	*pCtx;
{
    CX_CMD	*pCmdCtx=pCtx->pCmdCtx;
    long        stat;
    int         sigNum;
    char        clientName[30];
    static char sockPrompt[80];

/*-----------------------------------------------------------------------------
*       check to see if a connect attempt has been made.  If the server is
*       already in use by a client, connect attempts will be rejected.
*----------------------------------------------------------------------------*/
    if (ezsListenExclusiveUse(&pCtx->clientSock, &pCtx->listenSock,
				&pCtx->inUse, clientName, progName) < 0) {
	perror("check for connect attempt");
	exit(1);
    }

    if (pCtx->inUse && pCtx->inUseInit == 0) {
/*-----------------------------------------------------------------------------
*       When a connect attempt succeeds, change stdin and stdout to use
*       the socket, so that server I/O (i.e., I/O for this program) will
*       automatically be re-directed to the client.
*
*	The prompt is changed to a special prompt which ends with \n.  This
*	is done to avoid buffer flushing problems on some operating systems.
*	A special prefix of "#p#r#" is added to the normal prompt so that
*	the client knows that some massaging needs to be done before
*	displaying the prompt to the user.
*----------------------------------------------------------------------------*/
	AR_PRINTF "%s connected", clientName);
	if (arFreopenAllStd(pCtx, 1) != OK)
	    assertAlways(0);
	pCtx->oldPrompt = pCmdCtx->prompt;
	assert(strlen(pCtx->oldPrompt) < 72);
	(void)sprintf(sockPrompt, "#p#r#%s\n", pCtx->oldPrompt);
	pCmdCtx->prompt = sockPrompt;
	arListenSetReopen(pCtx, 2);	/* reopen all other tasks to socket */
	pCtx->inUseInit = 1;
	pCmdCtx->promptFlag = 1;
    }
    return OK;
}

/*+/subr**********************************************************************
* NAME	arListenClose - close a connection to a client
*
*-*/
long
arListenClose(pCtx)
AR_CTX	*pCtx;
{
    close(pCtx->clientSock);
    pCtx->clientSock = -1;
    pCtx->inUse = 0;
    pCtx->inUseInit = 0;
    if (arFreopenAllStd(pCtx, 0) != OK)
	assertAlways(0);
    AR_PRINTF "%s disconnected", pCtx->inUseName);
    pCtx->pCmdCtx->prompt = pCtx->oldPrompt;
    arListenSetReopen(pCtx, 0);	/* reopen fd's to std... */
    return OK;
}

/*+/subr**********************************************************************
* NAME	arListenQuit - quit listening for client connect attempts
*
*-*/
long
arListenQuit(pCtx)
AR_CTX	*pCtx;
{
    if (pCtx->clientSock != -1) {
        shutdown(pCtx->clientSock, 2);
        close(pCtx->clientSock);
    }

    close(pCtx->listenSock);
    return OK;
}

/*+/subr**********************************************************************
* NAME	arListenSetReopen - set flag to reopen std... fd's
*
*-*/
void
arListenSetReopen(pCtx, option)
AR_CTX	*pCtx;
int	option;		/* I option for reopen for use with arFreopenAllStd */
{
    pCtx->reopenIOoption = option;
    pCtx->reopenIO = 1;
}

/*+/subr**********************************************************************
* NAME	arFreopenAllStd - reopen stdin, stdout, and stderr
*
* DESCRIPTION
*
* RETURNS
*       OK, or
*       ERROR   (an error message has been printed)
*
*-*/
static int
arFreopenAllStd(pCtx, option)
AR_CTX *pCtx;		/* IO pointer to AR context */
int     option;		/* I activity selector:
			0  set back to use old fd's
			1  use client socket, save old fd's
			2  use client socket, don't save old fd's */
{
    if (option == 0) {
        if (ezsFreopenToFd(stdin, &pCtx->oldStdinFd, NULL) < 0) {
            perror("restore stdin");
            return ERROR;
        }
        if (ezsFreopenToFd(stdout, &pCtx->oldStdoutFd, NULL) < 0) {
            perror("restore stdout");
            return ERROR;
        }
        if (ezsFreopenToFd(stderr, &pCtx->oldStderrFd, NULL) < 0) {
            perror("restore stderr");
            return ERROR;
        }
    }    
    else if (option == 1) {
        if (ezsFreopenToFd(stdin, &pCtx->clientSock, &pCtx->oldStdinFd) < 0) {
            perror("reopen stdin");
            return ERROR;
        }
        if (ezsFreopenToFd(stdout, &pCtx->clientSock, &pCtx->oldStdoutFd) < 0) {
	    perror("reopen stdout");
            return ERROR;
        }
        if (ezsFreopenToFd(stderr, &pCtx->clientSock, &pCtx->oldStderrFd) < 0) {
	    perror("reopen stderr");
            return ERROR;
        }
    }
    else if (option == 2) {
        if (ezsFreopenToFd(stdin, &pCtx->clientSock, NULL) < 0) {
            perror("reopen stdin");
            return ERROR;
        }
        if (ezsFreopenToFd(stdout, &pCtx->clientSock, NULL) < 0) {
            perror("reopen stdout");
            return ERROR;
        }
        if (ezsFreopenToFd(stderr, &pCtx->clientSock, NULL) < 0) {
            perror("reopen stderr");
            return ERROR;
        }
    }    
    return OK;
}

/*+/subr**********************************************************************
* NAME	arInitAtStartup - initialization for AR
*
* DESCRIPTION
*	Perform several initialization duties:
*	o   initialize the AR context
*	o   initialize the command context block
*	o   initialize the help information
*
* RETURNS
*	OK, or
*	ERROR
*
*-*/
long
arInitAtStartup(pArCtx)
AR_CTX	*pArCtx;
{
    long	retStat=OK;
    long	stat;
    CX_CMD	*pCmdCtx=&pArCtx->cmdCtx;

    pArCtx->pReqHead = NULL;
    pArCtx->pReqTail = NULL;
    pArCtx->pCmdCtx = pCmdCtx;

    pArCtx->serverMode = 0;
    pArCtx->IPPort = 0;
    pArCtx->stop = 0;
    pArCtx->inUse = 0;
    pArCtx->inUseInit = 0;
    pArCtx->listenSock = -1;
    pArCtx->clientSock = -1;
    pArCtx->reopenIO = 0;
    pArCtx->oldStdinFd = -1;
    pArCtx->oldStdoutFd = -1;
    pArCtx->oldStderrFd = -1;

    cmdInitContext(pCmdCtx, "  AR_cmd: ");

/*-----------------------------------------------------------------------------
* help information initialization
*----------------------------------------------------------------------------*/
    helpInit(&pCmdCtx->helpList);
/*-----------------------------------------------------------------------------
* help info--legal generic commands
*----------------------------------------------------------------------------*/
    helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpCmds, "commands", "\n\
Generic commands are:\n\
   disconnect                (use    help usage   for more info)\n\
   help       [topic]\n\
   quit (or ^D)              (use    help usage   for more info)\n\
");
/*-----------------------------------------------------------------------------
* help info--legal AR-specific commands
*----------------------------------------------------------------------------*/
    helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpCmdsSpec, "commands", "\n\
AR-specific commands are:\n\
    activate    reqName\n\
    close       reqName  or  all\n\
    help        [topic]\n\
    inactivate  reqName\n\
    open        reqName  or  all\n\
    sample      reqName\n\
    startStepping reqName\n\
    status\n\
");
/*-----------------------------------------------------------------------------
* help info--AR usage information
*----------------------------------------------------------------------------*/
	helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpUsage, "usage", "\n\
AR_cmd (the data archiver) acquires data from Channel Access and saves the\n\
data to disk.  (The companion program ARR_cmd (archive retriever) can be\n\
used to retrieve the data from disk for printing, plotting, etc.)  AR and\n\
ARR are the corresponding GUI-based programs.\n\
\n\
AR_cmd can be run in interactive mode (or in a script) by:\n\
    % AR_cmd\n\
\n\
AR_cmd can be run as a server by:\n\
    % AR_cmd server &\n\
\n\
AR_cmd can be run as a client connecting to an AR_cmd server on \"hostName\" by:\n\
    % AR_cmd hostName\n\
\n\
    When running AR_cmd as a client, two commands are of special interest:\n\
\n\
    disconnect    shuts down the client but leaves the server running\n\
    quit (or ^D)  shuts down both the client and the server\n\
");
    return retStat;
}

#if 0
/*+/macro*********************************************************************
* NAME	arLogPrintf - do a printf to the archiver log
*
* SYNOPSIS
*	void
*	arLogPrintf(pCtx, fmt, ...)
*	AR_CTX	*pCtx;	/* I pointer to AR context */
*	char	*fmt;	/* I printf format string */
*	...		/* I other arguments, for printing */
*
*-*/
#endif
#ifdef __STDC__
void arLogPrintf(AR_CTX  *pCtx, char *fmt, ...)
{
#else
void arLogPrintf(va_alist)
va_dcl
{
    AR_CTX	 *pCtx;
    char	 *fmt;
#endif
    va_list	pArgs;
    char	message[500];

#ifdef __STDC__
    va_start(pArgs, fmt);
#else
    va_start(pArgs);
    pCtx = va_arg(pArgs, AR_CTX *);
    fmt = va_arg(pArgs, char *);
#endif
    vsprintf(message, fmt, pArgs);
    va_end(pArgs);
    arLogMessage(progName, message);
}

/*+/subr***********************************************************************
* NAME	arLogMessage - print a message and/or status on stderr
*
* DESCRIPTION
*	Prints a message on stderr, with a prefix of date and time and
*	program name.
*
*-*/
int
arLogMessage(progName, message)
char	*progName;	/* I program name (or " "; pglArCtx->progname
				assumed) */
char	*message;	/* I message to print, or NULL; don't end with \n */
{
    char	textTemp[80];	/* temp for text */
    TS_STAMP	timeTemp;	/* temp for present time */

    (void)tsLocalTime(&timeTemp);
    (void)tsStampToText(&timeTemp, TS_TEXT_MMDDYY, textTemp);
    if (message != NULL)
	(void)fprintf(stderr, "%.17s %-6s: %s\n", textTemp, progName, message);

    return OK;
}

/*+/subr**********************************************************************
* NAME	arReadEnabled - open everything in open.dat
*
*-*/
int
arReadEnabled(pCtx)
AR_CTX	*pCtx;
{
    FILE	*pStream;
    char	record[81];	/* record from file */
    char	file[80];
    char	req[80];
    int		i;

#define ENB_PATH "./arReq/open.dat"

    if ((pStream = fopen(ENB_PATH, "r")) == NULL) {
	AR_PRINTF "couldn't open request file: %s", ENB_PATH);
	return ERROR;
    }
    AR_PRINTF "**** reading: %s ****", ENB_PATH);
    while (fgets(record, 80, pStream) != NULL) {
	if ((i = strlen(record)) > 0) {
	    if (record[i-1] == '\n')
		record[i-1] = '\0';	/* remove \n before logging record */
	}
	if (strlen(record) > 0 && record[0] != '#') {
	    assert(AR_NAME_DIM == 36);
	    if (sscanf(record, "%35s", req) == 1) {
		arReqOpen(pCtx, req);
	    }
	}
    }
    (void)fclose(pStream);
    AR_PRINTF "**** EOF for %s ****", ENB_PATH);

    return OK;
}

/*+/subr**********************************************************************
* NAME	arReqClose
*
*-*/
arReqClose(pReq)
AR_REQ	*pReq;
{
    AR_CTX	*pCtx=pReq->pCtx;
    AR_CHAN	*pReqChan;
    TS_STAMP	nowTs;

    tsLocalTime(&nowTs);
    if (pReq->sample.type == REQ_TYPE_SNAP)
	arReqSampleSnapStop(pReq, nowTs);
    else
	arReqSampleMonStop(pReq, nowTs);

    reqTrigWrap(&pReq->actTrig);
    reqTrigWrap(&pReq->inactTrig);
    reqTrigWrap(&pReq->sample.forEvent);
    reqTrigWrap(&pReq->sample.trigger);
    reqStepWrap(&pReq->step);

    DoubleListRemove(pReq, pCtx->pReqHead, pCtx->pReqTail);
    for (; pReq->pChanHead!=NULL; ) {
	pReqChan = pReq->pChanHead;
	DoubleListRemove(pReqChan, pReq->pChanHead, pReq->pChanTail);
	if (pReqChan->pCfChan != NULL)
	    arCFChanClose(&pReqChan->pCfChan);
	if (pReq->pSspec != NULL && pReqChan->pSChan != NULL)
	sydChanClose(pReq->pSspec, pReqChan->pSChan);
	free((char *)pReqChan);
    }
    if (pReq->pSspec != NULL)
	sydClose(pReq->pSspec);
    if (pReq->pCfDesc != NULL)
	arCFClose(&pReq->pCfDesc);
    unlink(pReq->lockPath);
}

/*+/subr**********************************************************************
* NAME	arReqOpen - read an archive request file and open the request
*
*-*/
int
arReqOpen(pCtx, name)
AR_CTX	*pCtx;
char	*name;		/* name of request */
{
    RAW_REQ	*pRaw;
    char	file[250];		/* path for request file */
    AR_REQ	*pReq;
    FILE	*pFile;
    long	stat;
    enum sydChanSync sync;
    RAW_CHAN	*pChan;
    SYD_CHAN	*pSChan;
    AR_CHAN_DESC *pCfChan;
    AR_CHAN	*pReqChan;

    if (arReqFind(pCtx, name) != NULL) {
	printf("...%s already opened\n", name);
	return ERROR;
    }

    sprintf(file, "%s.req", name);

    pRaw = rawReqReadFile("./arReq", file, NULL, NULL);
    if (pRaw == NULL || pRaw->totalErrors > 0) {
	printf("...errors encountered reading %s--aborting open\n", file);
	if (pRaw != NULL)
	    rawReqFree(pRaw);
	return ERROR;
    }

    pReq = (AR_REQ *)malloc(sizeof(AR_REQ));
    assertAlways(pReq != NULL);
    strcpy(pReq->name, pRaw->name);
    strcpy(pReq->dir, pRaw->dir);
    strcpy(pReq->file, pRaw->file);
    sprintf(pReq->lockPath, "%s.open", file);
    if (arLock(pCtx, pReq->name, pReq->lockPath, 1) != 1) {
	rawReqFree(pRaw);
	free((char *)pReq);
	return;
    }
    pReq->pCtx = pCtx;
    pReq->pCfDesc = NULL;
    pReq->pChanHead = pReq->pChanTail = NULL;
    pReq->pNext = pReq->pNext = NULL;
    pReq->sample = pRaw->sample;
    pReq->actTrig = pRaw->actTrig;
    pReq->allowDiscon = pRaw->allowDiscon;
    pReq->inactTrig = pRaw->inactTrig;
    pReq->pSspec = NULL;
    pReq->status = REQ_ST_ENABLED;
    pReq->nDis = pReq->nCtrlDis = -1;
    pReq->sampling = 0;
    pReq->step = pRaw->step;
    pReq->getGR = 0;

    if (pReq->sample.outputType == REQ_OUT_BCF) {
	FILE	*pFile;
	char	path[240];

	sprintf(path, "%s/%s", pReq->sample.outDir, pReq->sample.outFile);
	if ((pFile = fopen(path, "r")) == NULL) {
	    pReq->pCfDesc = arCFCreate(path, pReq->sample.outSize);
	    if (pReq->pCfDesc == NULL) {
		printf("couldn't create file %s\n", path);
		perror(NULL);
		rawReqFree(pRaw);
		unlink(pReq->lockPath);
		free((char *)pReq);
		return;
	    }
	}
	else {
	    fclose(pFile);
	    pReq->pCfDesc = arCFOpen(path, O_RDWR);
	    if (pReq->pCfDesc == NULL) {
		printf("couldn't open file %s\n", path);
		perror(NULL);
		rawReqFree(pRaw);
		unlink(pReq->lockPath);
		free((char *)pReq);
		return;
	    }
	}
    }

    DoubleListAppend(pReq, pCtx->pReqHead, pCtx->pReqTail);
    reqTrigInit(&pReq->actTrig);
    reqTrigInit(&pReq->inactTrig);
    reqTrigInit(&pReq->sample.forEvent);
    reqTrigInit(&pReq->sample.trigger);
    reqStepInit(&pReq->step);

    stat = sydOpenCA(&pReq->pSspec, NULL);
    assertAlways(stat == S_syd_OK);
    pReq->pSspec->roundNsec = pRaw->sample.roundMsec*1000000;
    if (pReq->sample.type == REQ_TYPE_MONITOR)
	sydSetAttr(pReq->pSspec, SYD_ATTR_MON_FN, 0, arReqSampleMonHandler);
    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext) {
	pReqChan = (AR_CHAN *)malloc(sizeof(AR_CHAN));
	assertAlways(pReqChan != NULL);
	pReqChan->pReq = pReq;
	pReqChan->pCfChan = NULL;
	pReqChan->pSChan = NULL;
	if (pChan->filled)	sync = SYD_SY_FILLED;
	else			sync = SYD_SY_NONF;
	stat = sydChanOpen(pReq->pSspec, &pSChan, pChan->name,sync,pReqChan,0);
	assertAlways(stat == S_syd_chanNotConn);
	if (pReq->sample.outputType == REQ_OUT_BCF) {
	    pCfChan = arCFChanOpen(pReq->pCfDesc, pChan->name, O_RDWR);
	    if (pCfChan == NULL) {
		printf("couldn't do arCFChanOpen for %s", pChan->name);
		free((char *)pReqChan);
		rawReqFree(pRaw);
		return;
	    }
	    pReqChan->pCfChan = pCfChan;
	    pReqChan->pSChan = pSChan;
	    if ((pCfChan->flags & AR_CDESC_NEW) != 0)	pReqChan->getGR = 1;
	    else					pReqChan->getGR = 0;
	    if (pReqChan->getGR)
		pReq->getGR++;
	    DoubleListAppend(pReqChan, pReq->pChanHead, pReq->pChanTail);
	}
    }
    AR_PRINTF "%s opened", pReq->name);
    rawReqFree(pRaw);
}

/*+/subr**********************************************************************
* NAME	arReqFind - search for named request in open list
*
*-*/
AR_REQ *
arReqFind(pCtx, name)
AR_CTX	*pCtx;
char	*name;		/* name of request */
{
    AR_REQ	*pReq;
    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
	if (strcmp(pReq->name, name) == 0)
	    return pReq;
    }
    return NULL;
}

/*+/subr**********************************************************************
* NAME	arReqSampleMonHandler - handle Channel Access monitors for a request
*
*-*/
void arReqSampleMonHandler(arg)
struct event_handler_args arg;
{
    AR_CHAN	*pChan=(AR_CHAN *)arg.usr;
    AR_REQ 	*pReq=pChan->pReq;
    long	stat;
    TS_STAMP	*pStamp;

    if (pReq->sampling == 0)
	return;
    pStamp = &((struct dbr_time_float *)(arg.dbr))->stamp;
    if (pReq->pSspec->roundNsec > 0)
	sydTsRound(pStamp, pReq->pSspec->roundNsec);
    if (pReq->firstTs.secPastEpoch == 0)
	pReq->firstTs = *pStamp;
    if (TsCmpStampsGT(pStamp, &pReq->lastTs))
	pReq->lastTs = *pStamp;
    stat = arCFChanWrite(pChan->pCfChan, &arg);
    assert(stat == OK);
    pReq->monitorCount++;
}

/*+/subr**********************************************************************
* NAME	arReqSampleMonStart - start sampling in monitor mode
*
*-*/
arReqSampleMonStart(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;
    AR_CHAN	*pRChan;

    if (pReq->status != REQ_ST_ACTIVE) {
	AR_PRINTF "%s: can't monitor--state isn't `active'", pReq->name);
	return;
    }
    AR_PRINTF "%s: monitor sampling...", pReq->name);
    pReq->sample.halt = 0;
    pReq->sampling = 1;
    pReq->firstTs.secPastEpoch = pReq->lastTs.secPastEpoch = 0;
    pReq->firstTs.nsec = pReq->lastTs.nsec = 0;
    pReq->monitorCount = 0;
    if (pReq->sample.forType == REQ_TYPE_TIMED)
	TsAddStamps(&pReq->sample.forStop, &nowTs, &pReq->sample.forStamp);

    if (pReq->step.samplePending) {
	pReq->step.samplePending = 0;
	pReq->step.sampleActive = 1;
    }
    for (pRChan=pReq->pChanHead; pRChan!=NULL; pRChan=pRChan->pNext)
	arCFChanWriteByte(pRChan->pCfChan, AR_DAT_IC_MON_BEGIN);
    sydMonitorStart(pReq->pSspec);
}

/*+/internal******************************************************************
* NAME	arReqSampleMonStop - stop sampling in monitor mode
*
*-*/
arReqSampleMonStop(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    char	fromTxt[28], toTxt[28];

    if (pReq->sampling == 0)
	return;
    sydMonitorStop(pReq->pSspec);
    pReq->sampling = 0;
    if (pReq->sample.halt)
	AR_PRINTF "%s: monitoring halted; ", pReq->name);
    else
	AR_PRINTF "%s: monitoring stopped; ", pReq->name);
    reqTrig_setDone(&pReq->inactTrig);
    tsStampToText(&pReq->firstTs, TS_TEXT_MMDDYY, fromTxt);
    tsStampToText(&pReq->lastTs, TS_TEXT_MMDDYY, toTxt);
    AR_PRINTF "%s: %d values from %s to %s", pReq->name,
			    pReq->monitorCount, fromTxt, toTxt);
    stat = arCFSyncWrite(pReq->pCfDesc);
    assert(stat == OK);
}

/*+/internal******************************************************************
* NAME	arReqSampleSnapStart - start sampling in snapshot mode
*
*-*/
arReqSampleSnapStart(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    int		nCon, nDis;
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;

    if (pReq->status != REQ_ST_ACTIVE) {
	AR_PRINTF "%s: can't snapshot--state isn't `active'", pReq->name);
	return;
    }
    if (pReq->sample.trigger.eventHappened ||
				pReq->sample.startWhen != REQ_TYPE_BEFORE) {
	AR_PRINTF "%s: snapshot sampling...", pReq->name);
	if (pReq->sample.forType == REQ_TYPE_TIMED)
	    TsAddStamps(&pReq->sample.forStop, &nowTs, &pReq->sample.forStamp);
	pReq->sampling = 1;
    }
    else {
	AR_PRINTF "%s: queueing pre-trigger samples...", pReq->name);
	pReq->sampling = -1;
    }
    pReq->sample.halt = 0;
    if (pReq->step.samplePending) {
	pReq->step.samplePending = 0;
	pReq->step.sampleActive = 1;
    }

    sydInputResetKeepNewest(pSspec);
    sydSampleSetAlloc(pSspec, pReq->sample.bufNsamples);
    pSspec->refTs.secPastEpoch = 0;
    pSspec->refTs.nsec = 0;
    if (pReq->sample.startWhen != REQ_TYPE_BEFORE) {
	if (pReq->sample.forType == REQ_TYPE_TIMED)
	    TsAddStamps(&pReq->sample.forStop, &nowTs, &pReq->sample.forStamp);
    }
}

/*+/internal******************************************************************
* NAME	arReqSampleSnapStop - stop sampling in snapshot mode
*
*-*/
arReqSampleSnapStop(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;
    char	fromTxt[28], toTxt[28];
    chtype	type;
    AR_CHAN	*pRChan;
    void	*pData;

    if (pReq->sampling != 1)
	return;
    pReq->sampling = 0;
    if (pReq->sample.halt) {
	AR_PRINTF "%s: snapshot aborted", pReq->name);
	return;
    }
    reqTrig_setDone(&pReq->inactTrig);
    if (pSspec->sampleCount <= 0) {
	AR_PRINTF "%s: snapshot complete; no samples retrieved", pReq->name);
	return;
    }
    tsStampToText(&pSspec->pTimeStamp[pSspec->firstData],
			    TS_TEXT_MMDDYY, fromTxt);
    tsStampToText(&pSspec->pTimeStamp[pSspec->lastData],
			    TS_TEXT_MMDDYY, toTxt);
    AR_PRINTF "%s: snapshot complete; %d samples from %s to %s", pReq->name,
			    pSspec->sampleCount, fromTxt, toTxt);
    if (pReq->sample.outputType == REQ_OUT_SSF) {
	FILE	*pFile;
	char	path[240];

	sprintf(path, "%s/%s", pReq->sample.outDir, pReq->sample.outFile);
	if ((pFile = fopen(path, "a")) == NULL) {
	    AR_PRINTF "%s: snapshot aborted--couldn't open %s",pReq->name,path);
	    return;
	}
	sydSampleSetWriteSSF(pReq->pSspec, pFile,
			"PROGRAM: AR", "operator snapshot");
	fclose(pFile);
    }
    else {
	int		samp, count;
	SYD_CHAN	*pSChan;

	for (pRChan=pReq->pChanHead; pRChan!=NULL; pRChan=pRChan->pNext) {
	    pSChan = pRChan->pSChan;
	    type = pSChan->dbfType;
	    count = pSChan->elCount;
	    samp = pSspec->firstData;
	    if (pSChan->pData == NULL)
		continue;	/* skip to end of for loop */
	    while (1) {
		if (pSChan->pFlags[samp].eof)
		    break;
		if (pSChan->pFlags[samp].missing)
		    ;	/* no action on missing data */
		else if (samp != pSspec->firstData &&
					pSChan->pDataCodeL[samp] == 'F') {
		    ;	/* no action on filled data, except first */

#if 0
jjj	/* need to handle snap_begin and _end for filled stuff */
	/* also need to consider what to do if start a snap write and
		it's interrupted by EOF or INACT */
#endif

		}
		else {
		    if (samp == pSspec->firstData) {
			if (pSspec->sampleCount > 1) {
			    arCFChanWriteByte(pRChan->pCfChan,
						AR_DAT_IC_SNAP_BEGIN);
			}
			else {
			    arCFChanWriteByte(pRChan->pCfChan,
						AR_DAT_IC_SNAP_SINGLE);
			}
		    }
		    else if (samp == pSspec->lastData)
			arCFChanWriteByte(pRChan->pCfChan, AR_DAT_IC_SNAP_END);
		    pData = (void *)(((char *)pSChan->pData) +
					samp * dbr_value_size[type]);
		    stat = arCFChanWrite_args(pRChan->pCfChan,
			    pSspec->pTimeStamp[samp], type, count, pData,
			    pSChan->pDataAlStat[samp],
			    pSChan->pDataAlSev[samp]);
		    assert(stat == OK);
		}
		if (samp == pSspec->lastData)
		    break;
		else if (++samp >= pSspec->dataDim)
		    samp = 0;
	    }
	}
	stat = arCFSyncWrite(pReq->pCfDesc);
	assert(stat == OK);
    }
}

/*+/subr**********************************************************************
* NAME	arReqStatus
*
*-*/
arReqStatus(pReq, type)
AR_REQ	*pReq;
AR_INF	type;
{
    int		nCon, nDis, nCtrlDis;
    AR_CTX	*pCtx=pReq->pCtx;
    char	lastTxt[28];
    SYD_SPEC	*pSspec=pReq->pSspec;

    arReqCheckConn(pReq, &nCon, &nDis, &nCtrlDis, 0);
    if (type == AR_INF_STATE) {
	if (pReq->status == REQ_ST_ENABLED) {
	    AR_PRINTF "   opened; %d connected, %d connections pending",
			nCon, nDis);
	}
	else if (pReq->status == REQ_ST_ACTIVE)
	    AR_PRINTF "   activated");
	else if (pReq->status == REQ_ST_INACTIVE) {
	    AR_PRINTF "   inactivated; %d connected, %d connections pending",
			nCon, nDis);
	}
    }
    else if (type == AR_INF_CONN) {
	AR_PRINTF "   %d connected, %d connections pending", nCon, nDis);
    }
    else if (type == AR_INF_SAMP) {
	if (pReq->sampling) {
	    if (pReq->sample.type == REQ_TYPE_SNAP) {
		if (pReq->sampling == 1) {
		    tsStampToText(&pSspec->pTimeStamp[pSspec->lastData],
			    TS_TEXT_MMDDYY, lastTxt);
		    AR_PRINTF "   snapshot in progress; %d samples, last at %s",
			    pSspec->sampleCount, lastTxt);
		}
		else {
		    AR_PRINTF "   pre-trigger sampling");
		}
	    }
	    else {
		tsStampToText(&pReq->lastTs, TS_TEXT_MMDDYY, lastTxt);
		AR_PRINTF "   monitoring in progress; %d values, last at %s",
			    pReq->monitorCount, lastTxt);
	    }
	}
	else
	    AR_PRINTF "   not presently sampling");
    }
    else if (type == AR_INF_STEP) {
	char	*pState;
	if (pReq->step.active) {
	    if (pReq->step.getPending)		pState = "get pending";
	    else if (pReq->step.delayPending)	pState = "delay pending";
	    else if (pReq->step.samplePending)	pState = "sample pending";
	    AR_PRINTF "   stepping active: %s", pState);
	    AR_PRINTF "      last val sent=%f", pReq->step.present);
	    if (pReq->step.complete)
		AR_PRINTF "      sequence is complete");
	}
	else
	    AR_PRINTF "   not stepping");
    }
}

/*+/subr**********************************************************************
* NAME	arLock - provide a "global" lock and unlock mechanism
*
* DESCRIPTION
*
* RETURNS
*	1	if lock succeeds
*	0	if unlock succeeds
*	-1	if either operation fails
*
*-*/
int
arLock(pCtx, text, lockPath, lock)
AR_CTX	*pCtx;
char	*text;		/* I name or description for error messages */
char	*lockPath;	/* I pointer to path for lock file */
int	lock;		/* I 0,1 to reset,set the lock */
{
    FILE	*pFile;
    char	msgLock[240];
    char	msgWho[240];
    TS_STAMP	ts;
    char	tsText[32];
    int		c;

    if (lock) {
	if ((pFile = fopen(lockPath, "r+")) != NULL) {
	    fread(msgWho, 240, 1, pFile);
	    fclose(pFile);
	    if (strncmp(msgWho, "unlocked", 8) != 0) {
		sprintf(msgLock, "%s is locked:", text);
		printf("%s %s\n", msgLock, msgWho);
		if (pCtx->serverMode)
		    return -1;
		printf("if you wish to override the lock type the letter o: ");
		fflush(stdout);
		c = fgetc(stdin);
		if (c != 'o' && c != 'O')
		    return -1;
	    }
	}
    }
    if ((pFile = fopen(lockPath, "w+")) == NULL) {
	perror("couldn't open lock file");
	return -1;
    }
    if (lock) {
	char	*host[20];
	tsLocalTime(&ts);
	tsStampToText(&ts, TS_TEXT_MONDDYYYY, tsText);
	gethostname(host, 20);
	sprintf(msgWho, "by:%s  on:%s  at:%.21s",
			getenv("LOGNAME"), host, tsText);
	fwrite(msgWho, strlen(msgWho)+1, 1, pFile);
	fclose(pFile);
	return 1;
    }
    else {
	fwrite("unlocked", 9, 1, pFile);
	fclose(pFile);
	unlink(lockPath);
	return 0;
    }
}
