/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/*	@(#)arr_tk.c	1.0 1/04/94
 *	Author:	Carl A. Timmer
 *	Date:	01-04-94
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 * make options
 *	-DNDEBUG	don't compile assert() checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */

/****************************************************************************
* TITLE	arr_tk - archive data retrieval routine to be used with tcl/tk
*
* DESCRIPTION
*
*	Most of this code was copied from Roger Cole's ARR_cmd.c in order
*	to provide a tcl/tk interface to the EPICS archiving software.
*	For more details, see "Archiver Retrieval Tool: arr Programmer's 
*	Guide" or the user's guide.
*
*****************************************************************************/

/* #define DB_TEXT_GLBLSOURCE*/	/* need to get db_access definitions */

#include <stdio.h>
#include <stdlib.h>		/* need for "atof" definition */
#include <sys/file.h>		/* for O_RDWR and O_RDONLY definitions */
#include <string.h>
#ifndef SOLARIS
#include <strings.h>		/* for SSF fileInfo func */
#endif
#include <ctype.h>
#include <math.h>
#include <signal.h>
#include <setjmp.h>

#include <X11/Xlib.h>
#include <genDefs.h>
#include <cmdDefs.h>
#include <tsDefs.h>
#include <sydDefs.h>
#include <alarm.h>
#include <arCS.h>
#include <cadef.h>
#include <db_access.h>
#include <arAccessLib.h>	/* for CF fileInfo func */
#include <limits.h>

#include <tcl.h>		/* for tcl definitions */


/*---------------------------------------------------------------------------
* ARR_CTX - context information
*
*	An ARR descriptor is the `master handle' which is used for
*	handling business.
*----------------------------------------------------------------------------*/

typedef struct ARRCtx {
    CX_CMD	cmdCtx;
    CX_CMD	*pCmdCtx;
    jmp_buf	sigEnv;
    char	bychanFile[80];		/* by channel data file name */
    char	sampleFile[80];		/* sample set data file name */
    SYD_SPEC	*pSspec;		/* synchronous set spec */
    int		useStats;		/* 1 for work with mean and stdDev */
    int		useMeans;		/* 1 for work with mean */
    int		accept;			/* 1 if restricting acceptance */
    char	acceptText[120];	/* text for accept condition */
    int		deadRst;		/* 1 for reset fill after deadtime */
    TS_STAMP	deadRefTs;		/* reference stamp for deadtime */
    double	deadRstSec;		/* deadtime restart interval, in sec */
    int		discardPartial;		/* 1 to discard partial samples */
    int		stop;			/* 1 says `stop operations' */

/* Below added by C. Timmer */
    Tcl_DString	returnValue;		/* tcl list data returned for "arr_getData" */
    TS_STAMP	startingTS;		/* earliest TS with which to start data load */
    TS_STAMP	endingTS;		/* latest   TS with which to end   data load */
    int		numberSamples;		/* number of data samples to collect */
    char	retrievalCode[10];	/* if = TIMESTAMP, the dataload end condition
					   is the time given by "endingTS";
					   if = NUBMER, dataload ending condition is the
					   # of data samples given by "numberSamples" */
} ARR_CTX;

/*-----------------------------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
int Arr_FileList _ANSI_ARGS_((ClientData clientData,
				Tcl_Interp *interp, int argc, char **argv));
int Arr_GetData _ANSI_ARGS_((ClientData clientData,
				Tcl_Interp *interp, int argc, char **argv));
int Arr_TsAddSec _ANSI_ARGS_((ClientData clientData,
				Tcl_Interp *interp, int argc, char **argv));
int Arr_TsCmpStamps _ANSI_ARGS_((ClientData clientData,
				Tcl_Interp *interp, int argc, char **argv));
int Arr_DiffAsSec _ANSI_ARGS_((ClientData clientData,
				Tcl_Interp *interp, int argc, char **argv));
int Arr_DiffAsStamp _ANSI_ARGS_((ClientData clientData,
				Tcl_Interp *interp, int argc, char **argv));

static int  arrCmdCrunch _ANSI_ARGS_((ARR_CTX *pArrCtx, Tcl_Interp *interp));
static long arrInitAtStartup _ANSI_ARGS_((ARR_CTX *pArrCtx));
static void arrSigHandler _ANSI_ARGS_((int signo));

static void ARR_channel _ANSI_ARGS_((ARR_CTX *pArrCtx));
static void ARR_delete _ANSI_ARGS_((ARR_CTX *pArrCtx));
static void ARR_position _ANSI_ARGS_((ARR_CTX *pArrCtx));
static int  ARR_sample _ANSI_ARGS_((ARR_CTX *pArrCtx, int reqCount));

static int  ARR_getRegularData _ANSI_ARGS_((ARR_CTX *pArrCtx, Tcl_Interp *interp));
static int  ARR_getStatData _ANSI_ARGS_((ARR_CTX *pArrCtx, Tcl_Interp *interp));
static int  ARR_getData _ANSI_ARGS_((ARR_CTX *pArrCtx, Tcl_Interp *interp));

/**************************************************************************
* NAME	posXxx - text file search and positioning routines
***************************************************************************/
/* The following are copied from sydSubrSSF.c */
static long posBOF _ANSI_ARGS_((FILE *pStream));
static long posEOF _ANSI_ARGS_((FILE *pStream));
static long posKeyFwd _ANSI_ARGS_((char *key, FILE *pStream));
static long posKeyRev _ANSI_ARGS_((char *key, FILE *pStream, long lastPos));
static long posPos _ANSI_ARGS_((FILE *pStream, long pos));
static long sydSSFFuncSeekStamp _ANSI_ARGS_((FILE *ssFile,
				TS_STAMP *pStamp, int revFlag, long pos));

/*-----------------------------------------------------------------------------
* definitions
*----------------------------------------------------------------------------*/

static ARR_CTX	*pglArrCtx;

#define SHOW_AR 4		/* from sydSubr.c */
#define P_CMD	pCmdCtx->pCommand
#define P_FLD	pCmdCtx->pField
#define FLD_LEN	pCmdCtx->fldLen
#define P_LINE	pCmdCtx->pLine
#define DELIM	pCmdCtx->delim


/******************************************************************************
* NAME	Arr_FileList
*
* DESCRIPTION
*	This here subroutine is a so-called "command procedure" in "tcl".
*	It follows the tcl protocol in eventually establishing a tcl command,
*	"arr_fileList", which returns a list of all pvnames in an archiver-
*	generated file. The first input argument is the file name and the
*	second is the path name.
*
*	This subroutine "becomes" a tcl command by being declared in the file,
*	~tcl_home/tcl7.3/tclAppInit.c  using Tcl_CreateCommand.
*
********************************************************************************/
#ifdef __STDC__
int
Arr_FileList(
  ClientData	clientData,
  Tcl_Interp	*interp,
  int		argc,
  char		**argv)
#else
int
Arr_FileList(clientData, interp, argc, argv)
  ClientData	clientData;
  Tcl_Interp	*interp;
  int		argc;
  char		**argv;
#endif /*__STDC__*/

{
	/* definitions - general */
	int		code;
	char		openFileCommandPrefix[10],
			openFileCommand[100],
			filename[80],
			line[50];
	FILE		*fd;
	Tcl_DString pvNames;	/* tcl dynamics strings, used to hold channel names */
	

	/* definitions for reading of both 
	   "byChannel" & "sampleSet" files */
	int		j;
	TS_STAMP	oldestStamp,
			newestStamp,
			ts1, ts2;
	char		oldestStampText[28],
			newestStampText[28],
			chanName[db_name_dim+2];

	/* definitions for reading of "sampleSet" files */
	FILE		*ssFile;
	char		record[120],
			*pRecord,
			*pField,
			delim,
			recDelim;
	int		elCount;

	/* definitions for reading of "byChannel" files */
	AR_CF_DESC	*pCfDesc;	/* pointer to cfDesc for channel file */

	/* definitions for archive commands */
	long	stat;		/* status return from calls */
	ARR_CTX	arrCtx, *pCtx=&arrCtx;

	/* Initialize Tcl dynamic string */
	Tcl_DStringInit(&pvNames);

/************************/
/* check # of arguments */
/************************/
	if(argc != 3)
	{
		/* send tcl error message */
		Tcl_SetResult(interp, "arr_fileList: wrong # args", TCL_STATIC);
		return TCL_ERROR;
	}

/**************************/
/* check file name & type */
/**************************/

	/* The basic idea is to guess what type of file is being */
	/* accessed. Eliminate the need for the user to specify  */
	/* "byChannel" or "sampleSet".				 */

	sprintf(filename, "%s/%s", argv[1], argv[2]);

	/* check to see if the filename is a regular file using tcl command */
	code = Tcl_VarEval(interp, "file isfile ", filename, (char *) NULL);
	if(code != TCL_OK)
	{
		Tcl_SetResult(interp, "arr_fileList: tcl command problems", TCL_STATIC);
		return TCL_ERROR;
	}

	if(strcmp(interp->result, "1") != 0)
	{
		Tcl_SetResult(interp, "arr_fileList: Select a regular file!", TCL_STATIC);
		return TCL_ERROR;
	}


	/* check file suffix for ".bin" or ".txt" extensions */
	/* to see if it's a "byChannel" or "sampleSet" file  */
	code = Tcl_VarEval(interp, "file extension ", argv[2], (char *) NULL);
	if(code != TCL_OK)
	{
		Tcl_SetResult(interp, "arr_fileList: tcl command problems", TCL_STATIC);
		return TCL_ERROR;
	}

	if(strcmp(interp->result, ".bin") == 0)
		strcpy(openFileCommandPrefix, "byChannel");

	else if(strcmp(interp->result, ".txt") == 0)
		strcpy(openFileCommandPrefix, "sampleSet");

	/* if no recognizable extension, see if the file */
	/* path is in an "arChan" or "arSet" directory   */
	else
	{
		code = Tcl_VarEval(interp,"file tail ", argv[1], (char *) NULL);
		if(code != TCL_OK)
		{
			Tcl_SetResult(interp, "arr_fileList: tcl command problems", TCL_STATIC);
			return TCL_ERROR;
		}

		if(strcmp(interp->result, "arChan") == 0)
			strcpy(openFileCommandPrefix, "byChannel");

		else if(strcmp(interp->result, "arSet") == 0)
			strcpy(openFileCommandPrefix, "sampleSet");

		/* If there is no recognizable extension */
		/* AND the file is in an unknown type of */
		/* directory, scan first 49 chars for \n.*/
		/* If there is one, it's probably ascii. */
		else
		{
			if((fd = fopen(filename, "r")) == NULL)
			{
			   Tcl_SetResult(interp, "arr_fileList: can't open file", TCL_STATIC);
			   return TCL_ERROR;
			}

			/* if "\n" found in first 49 chars, it's sampleSet */
			if(strchr(fgets(line, 50, fd) ,'\n') != NULL)
			    strcpy(openFileCommandPrefix, "sampleSet");
			else
			    strcpy(openFileCommandPrefix, "byChannel");

			fclose(fd);
		}
	}

	/* Clear the interp's result */
	Tcl_ResetResult(interp);

	/* put string into proper format for the ARR_cmd command cruncher */
	sprintf(openFileCommand, "%s %s", openFileCommandPrefix, filename);

/*****************************/
/* initialize archiver stuff */
/*****************************/

	arrInitAtStartup(pCtx);
	genSigInit(arrSigHandler);

	if (setjmp(pCtx->sigEnv) != 0)
	{
		
		/* wrap up loose ends before returning error */
		if (pCtx->pSspec != NULL)
		{
			stat = sydClose(pCtx->pSspec);
			assert(stat == S_syd_OK);
		}

		while (pCtx->pCmdCtx->pPrev != NULL)
			cmdCloseContext(&pCtx->pCmdCtx);

		if (pCtx->pCmdCtx->dataOut != stdout)
			fclose(pCtx->pCmdCtx->dataOut);

		Tcl_SetResult(interp, "arr_fileList: Problem initializing archiving routine", TCL_STATIC);
		return TCL_ERROR;
	}

/*************************************************/
/* use ARR_cmd's command cruncher to do the work */
/*************************************************/

    if (pCtx->stop == 0)
    {
	/* command to open archive file */
	strcpy(pCtx->pCmdCtx->line, openFileCommand);
	pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	stat = arrCmdCrunch(pCtx, interp);

	/* If the file doesn't open properly return error */
	if(stat == TCL_ERROR)
	{
	    Tcl_SetResult(interp, "arr_fileList: Problem opening archive file", TCL_STATIC);
	    return TCL_ERROR;
	}


	/* command to quit */
	strcpy(pCtx->pCmdCtx->line, "quit");
	pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	stat = arrCmdCrunch(pCtx, interp);
	if(stat == TCL_ERROR)
	{
	    Tcl_SetResult(interp, "arr_fileList: error quitting", TCL_STATIC);
	    return TCL_ERROR;
	}
    }

/*************************************/
/* use AR commands to read file info */
/*************************************/

    /* stole the following from sydSSFFunc (fileinfo command) */
    /* if file is "sample set" type ... */
    if(pCtx->pSspec->type == SYD_TY_SSF)
    {
    	if (pCtx->pSspec != NULL)
		ssFile = (FILE *)pCtx->pSspec->pHandle;

	if (posBOF(ssFile) < 0)
	{
		Tcl_SetResult(interp, "arr_fileList: error positioning file", TCL_STATIC);
		return TCL_ERROR;
	}

	if (sydSSFFuncSeekStamp(ssFile, &oldestStamp, 0, 0L) < 0)
	{
		Tcl_SetResult(interp, "arr_fileList: error finding oldest time stamp", TCL_STATIC);
		return TCL_ERROR;
	}

	if (sydSSFFuncSeekStamp(ssFile, &newestStamp, 1, -1L) < 0)
	{
		Tcl_SetResult(interp, "arr_fileList: error finding newest time stamp", TCL_STATIC);
		return TCL_ERROR;
	}

/* Find out how many time stamps there are
{
int i=0;
long pos= -1;

while ((pos = sydSSFFuncSeekStamp(ssFile, &ts1, 1, pos)) >= 0L)
	i++;
printf("Number of timestamps = %d\n", i);
}
*/

/*	printf("data from %s through %s\n",
	       tsStampToText(&oldestStamp, TS_TEXT_MMDDYY, oldestStampText),
               tsStampToText(&newestStamp, TS_TEXT_MMDDYY, newestStampText) );
*/
	tsStampToText(&oldestStamp, TS_TEXT_MMDDYY, oldestStampText);
        tsStampToText(&newestStamp, TS_TEXT_MMDDYY, newestStampText);

	if (strcmp(oldestStampText, "00:00:00.000") == 0 ||
		strcmp(oldestStampText, "00/00/00 00:00:00.000") == 0)
	{
		Tcl_SetResult(interp, "arr_fileList: No valid time stamps in file", TCL_STATIC);
		return TCL_ERROR;
	}

	if (posKeyRev("%endHeader%", ssFile, -1L) < 0)
	{
		Tcl_SetResult(interp, "arr_fileList: error locating %%endHeader%% in file", TCL_STATIC);
		return TCL_ERROR;
	}

	/* skip over endHeader */
	fgets(record, 120, ssFile);

	while (1)
	{
	    if (fgets(record, 120, ssFile) == NULL)
	    {
		Tcl_SetResult(interp, "arr_fileList: error reading channel name", TCL_STATIC);
		return TCL_ERROR;
	    }

	    pRecord = record;
	    recDelim = record[strlen(record)-1];

	    if (nextChanNameField(&pRecord, &pField, &delim) <= 1)
	    {
		Tcl_SetResult(interp, "arr_fileList: error reading channel type", TCL_STATIC);
		return TCL_ERROR;
	    }

	    if (strcmp(pField, "%endData%") == 0)
		break;

	    strcpy(chanName, pField);

	    if (nextAlphField(&pRecord, &pField, &delim) <= 1)
	    {
		Tcl_SetResult(interp, "arr_fileList: error reading channel type", TCL_STATIC);
		return TCL_ERROR;
	    }

	    if (sscanf(pRecord, "%*d %*d %d", &elCount) != 1)
	    {
		Tcl_SetResult(interp, "arr_fileList: error reading element count", TCL_STATIC);
		return TCL_ERROR;
	    }

	    j = strlen(chanName);
	    if (j > 4)
	    {
		if (strncmp(&chanName[j-4], ".VAL", 4) == 0)
		    chanName[j-4] = '\0';
	    }

	    /* Put each pvname onto the end of a sublist */
	    Tcl_DStringAppendElement(&pvNames, chanName);

	    /* printf("name = %s, count =  %d, type =  %s\n",chanName, elCount, pField); */
	    if (recDelim != '\n')
		while ((delim = fgetc(ssFile)) != EOF && delim != '\n')
		    ;	/* keep skipping until end of line */
	}

	if (posBOF(ssFile) < 0)
	{
		Tcl_SetResult(interp, "arr_fileList: error positioning file", TCL_STATIC);
		return TCL_ERROR;
	}
    }


    /* if file is "byChannel" type ... */
    else if(pCtx->pSspec->type == SYD_TY_CF)
    {
	if (pCtx->pSspec != NULL)
		pCfDesc = (AR_CF_DESC *)pCtx->pSspec->pHandle;

	assert(pCfDesc != NULL);
	if (ArB0MIHead(pCfDesc) != 0)
	{
	    AR_MI_BUF	*pMICurrent;
	    AR_CHAN_HDR	*pChanHdr;
	    int		iteration = 0;

	    assert(ArB0MIHead(pCfDesc) > 0);
	    assert(ArB0MITail(pCfDesc) > 0);
	    assert(ArB0MINblk(pCfDesc) > 0);
	    assert(pCfDesc->pMIHead != NULL);
	    assert(pCfDesc->pMITail != NULL);

	    pMICurrent = pCfDesc->pMIHead;
	    while (pMICurrent != NULL)
	    {
		int i;

		for (i=0; i<AR_CF_NHDR; i++)
		{
		    pChanHdr = &pMICurrent->chanHdr[i];

		    if (pChanHdr->name[0] != '\0')
		    {
			strcpy(chanName, pChanHdr->name);
			j = strlen(chanName);

			if (j > 4)
			{
			    if (strncmp(&chanName[j-4], ".VAL", 4) == 0)
				chanName[j-4] = '\0';
			}

			/* Put each pvname onto the end of a sublist */
			Tcl_DStringAppendElement(&pvNames, chanName);

			ts1.secPastEpoch = pChanHdr->oldestSecPastEpoch;
			ts1.nsec = 0;
			ts2.secPastEpoch = pChanHdr->newestSecPastEpoch;
			ts2.nsec = pChanHdr->newestNsec;

			if ((iteration == 0) || (TsCmpStampsGT(&oldestStamp, &ts1) == 1))
			{
			    oldestStamp.secPastEpoch = ts1.secPastEpoch;
			    oldestStamp.nsec = 0;
			    tsStampToText(&oldestStamp, TS_TEXT_MMDDYY,
				 oldestStampText);

			    if (strcmp(oldestStampText, "00:00:00.000") == 0 ||
			  	  strcmp(oldestStampText, "00/00/00 00:00:00.000") == 0)
			    {
				Tcl_SetResult(interp, "arr_fileList: Invalid time stamp in file", TCL_STATIC);
				return TCL_ERROR;
			    }

			}

			if ((iteration == 0) || (TsCmpStampsLT(&newestStamp, &ts2) == 1))
			{
			    newestStamp.secPastEpoch = ts2.secPastEpoch;
			    newestStamp.nsec = ts2.nsec;
			    tsStampToText(&newestStamp, TS_TEXT_MMDDYY,
				 newestStampText);

			    if (strcmp(newestStampText, "00:00:00.000") == 0 ||
			  	  strcmp(newestStampText, "00/00/00 00:00:00.000") == 0)
			    {
				Tcl_SetResult(interp, "arr_fileList: Invalid time stamp in file", TCL_STATIC);
				return TCL_ERROR;
			    }

			}

			iteration = 1;

			/* printf("CF: data from %s through %s\n",
			    oldestStampText, newestStampText); */
		    }
		}    
		pMICurrent = pMICurrent->pNextMI;
	    }
	}

	else
	{
		Tcl_SetResult(interp, "no channels in file", TCL_STATIC);
		return TCL_ERROR;
	}
    }

    else
	{ /* channel access or error */ }

/******************/
/* Wrap things up */
/******************/

    if (pCtx->pSspec != NULL)
    {
	stat = sydClose(pCtx->pSspec);
	assert(stat == S_syd_OK);
    }

    while (pCtx->pCmdCtx->pPrev != NULL)
	cmdCloseContext(&pCtx->pCmdCtx);

    if (pCtx->pCmdCtx->dataOut != stdout)
	fclose(pCtx->pCmdCtx->dataOut);

    /* put all data together in a list (of a list and individual values) */
    Tcl_DStringAppendElement(&pCtx->returnValue, Tcl_DStringValue(&pvNames));
    Tcl_DStringAppendElement(&pCtx->returnValue, oldestStampText);
    Tcl_DStringAppendElement(&pCtx->returnValue, newestStampText);
    Tcl_DStringAppendElement(&pCtx->returnValue, openFileCommandPrefix);

    /* Free up memory */
    Tcl_DStringFree(&pvNames);

    /* puts all retrieved data into result string for interp */
    Tcl_DStringResult(interp, &pCtx->returnValue);

   return TCL_OK;
}


/*******************************************************************************
* NAME	Arr_GetData
*
* DESCRIPTION
*	This here subroutine is a so-called "command procedure" in "tcl".
*	It follows the tcl protocol in eventually establishing a tcl command,
*	"arr_getData", which puts data into a number of global, tcl variables.
*	A list of these global variables get past to it in the arguments as 
*	well as a list of pvnames for which to get data from the file
*
*********************************************************************************/
#ifdef __STDC__
int
Arr_GetData(
  ClientData	clientData,
  Tcl_Interp	*interp,
  int		argc,
  char		**argv)
#else
int
Arr_GetData(clientData, interp, argc, argv)
  ClientData	clientData;
  Tcl_Interp	*interp;
  int		argc;
  char		**argv;
#endif /*__STDC__*/

{
	/* definitions - general */
	int	i,
		discardPartials_flag,
		mean_stdev_flag,
		num_millisec,
		num_samples,
		num_pvnames,
		num_items;

	char	openFileCommand[100],
		retrieveCode[10],
		accept_if[100],
		startingTimeStamp[28],
		filetype[10],
		filename[80],
		**pvnameList,
		**retrieveList,
		**pTScharArray;		/* Ptr to timestamp char array */
	
	/* definitions for archive commands */
	long	stat;		/* status return from calls */
	ARR_CTX	arrCtx, *pCtx=&arrCtx;

/************************/
/* check # of arguments */
/************************/
	if(argc > 11 || argc < 6)
	{
	    /* send tcl error message */
	    Tcl_SetResult(interp, "Useage: arr_getData (required arguments:)\n    filename\n    filetype\n    pvName_List\n    retrieveCode\n    retrieveList \n(other arguments optional:) \n    discard_partial_flag \n    mean_&_stdev_flag\n    round_timestamps_to\n    accept_if_condition", TCL_STATIC);
	    return TCL_ERROR;
	}

/*****************************/
/* initialize archiver stuff */
/*****************************/

	arrInitAtStartup(pCtx);
	genSigInit(arrSigHandler);

	if (setjmp(pCtx->sigEnv) != 0)
	{
		/* wrap up loose ends before returning error */
		if (pCtx->pSspec != NULL)
		{
			stat = sydClose(pCtx->pSspec);
			assert(stat == S_syd_OK);
		}

		while (pCtx->pCmdCtx->pPrev != NULL)
			cmdCloseContext(&pCtx->pCmdCtx);

		if (pCtx->pCmdCtx->dataOut != stdout)
			fclose(pCtx->pCmdCtx->dataOut);

		Tcl_SetResult(interp, "Problem initializing archiving routine", TCL_STATIC);
		return TCL_ERROR;
	}

	/* Clear the interp's result */
	Tcl_ResetResult(interp);

/*************************************************************************
 * Parse first 6 arguments (filename, filetype, pvList, retrieval code,  *
 * starting TS, ending TS or # of samples). These arguments are required.*
 *************************************************************************/

	/* read the file name */
	strcpy(filename, argv[1]);
/* printf("argv[1] = %s\n", filename); */

	/* read the file type */
	strcpy(filetype, argv[2]);
/* printf("argv[2] = %s\n", filetype); */
	if ((strcmp("byChannel", filetype) != 0)&&(strcmp("sampleSet", filetype) != 0))
	{
	    Tcl_SetResult(interp, "arr_getData: wrong filetype", TCL_STATIC);
	    return TCL_ERROR;
	}

	/* put string into proper format for the ARR_cmd command cruncher */
	sprintf(openFileCommand, "%s %s", filetype, filename);

	/* put the pvlist into char array & (int) num_pvnames */
	if (Tcl_SplitList(interp, argv[3], &num_pvnames, &pvnameList) != TCL_OK)
	    return TCL_ERROR;

	/* Put retrieval code into char array (retrieveCode):	*/
	/*  TIMESTAMP, beginning & ending timestamps		*/
	/*    NUMBER+, starting timestamp & #_of_samples forward  in time */
	/* or NUMBER-, starting timestamp & #_of_samples backward in time */
	strcpy(retrieveCode, argv[4]);
	strcpy(pCtx->retrievalCode, retrieveCode);
/* printf("argv[4] = %s\n", retrieveCode); */

	/****************************************************************
	 * Deconvolute the beginning and ending points for data loading	*
	 * from argv[5] which is a tcl list containing: { {earliestTS},	*
	 * {startingTS}, {endingTS or #ofSamples}}			*
	 ****************************************************************/

	/* put the retrieval list items into char array */
	if (Tcl_SplitList(interp, argv[5], &num_items, &retrieveList) != TCL_OK)
	    return TCL_ERROR;
/*
for (i=0;i<num_items;i++)
    printf("argv[5][%d] = %s\n", i, retrieveList[i]);
*/
	if (num_items != 3)
	{
	    Tcl_SetResult(interp, "arr_getData: wrong # items in 5th arg.", TCL_STATIC);
	    return TCL_ERROR;
	}

	/* put file's starting TS into structure; do strcpy FIRST! */
	strcpy(startingTimeStamp, retrieveList[1]);
	tsTextToStamp(&pCtx->startingTS, &retrieveList[1]);

	/* put #_of_samples into "num_samples" or put timestamp */
	/* into "timeStamp2" depending on retrieval code.	*/
	if (strncmp(retrieveCode, "NUMBER", 6) == 0)
	{
	    num_samples = atoi(retrieveList[2]);
	    if ((num_samples > 10000) || (num_samples < 1))
	    {
		Tcl_SetResult(interp, "arr_getData: 0 < #_of_samples <=10000", TCL_STATIC);
		return TCL_ERROR;
	    }

	    /* store the "# of samples actually desired" in structure */
	    pCtx->numberSamples = num_samples;

	    /*	If you want samples BEFORE the specified TS, things become alot
		more difficult because Cole's software can only search forward
		in time. Thus, one must start at the earliest TS, get all data,
		and then ferret out the desired portion of it.
	    */
	    if (strcmp(retrieveCode, "NUMBER-") == 0)
	    {
		/* have all the data read, from earliest TS with 10,000 samples */
		num_samples = 10000;
		strcpy(startingTimeStamp, retrieveList[0]);
	    }
	}

	else if (strcmp(retrieveCode, "TIMESTAMP") == 0)
	{
	    num_samples = 10000;
	    tsTextToStamp(&pCtx->endingTS, &retrieveList[2]);
	}

	else
	{
	    Tcl_SetResult(interp, "arr_getData: retrieveCode error", TCL_STATIC);
	    return TCL_ERROR;
	}
/* printf("startingTimeStamp = %s\n", startingTimeStamp); */

       /* free memory allocated by Tcl_SplitList */
       free(retrieveList);

/**********************************************************************************/
/* parse last 4 arguments (partial_samp_flag, mean_&_stdev_flag, round_Tstamp_to, */
/* accept_if_cond). These arguments are NOT required.				  */
/**********************************************************************************/

	if (argc >= 7)
	{
	    /* determine if we must discard partial samples */
	    discardPartials_flag = atoi(argv[6]);
/* printf("argv[6] = %d\n", discardPartials_flag); */
	    if (discardPartials_flag != 1 && discardPartials_flag != 0)
	    {
		Tcl_SetResult(interp, "arr_getData: \"discard_partial_flag\" must be 0 or 1", TCL_STATIC);
		return TCL_ERROR;
	    }
	}
	else
	    discardPartials_flag = 0;

	if (argc >= 8)
	{
	    /* determine if we want to use means & stdev */
	    mean_stdev_flag = atoi(argv[7]);
/* printf("argv[7] = %d\n", mean_stdev_flag); */
	    if (mean_stdev_flag != 1 && mean_stdev_flag != 0)
	    {
		Tcl_SetResult(interp, "arr_getData: \"mean_&_stdev_flag\" must be 0 or 1", TCL_STATIC);
		return TCL_ERROR;
	    }
	}
	else
	    mean_stdev_flag = 0;

	if (argc >= 9)
	{
	    /* round time stamps to the nearest x millisecond */
	    /* If num_millisec = 0, don't round timestamps.   */
	    num_millisec = atoi(argv[8]);
/* printf("argv[8] = %d\n", num_millisec); */
	    if (num_millisec < 0)
	    {
		Tcl_SetResult(interp, "arr_getData: \"round_timestamps_to\" must be >= 0", TCL_STATIC);
		return TCL_ERROR;
	    }
	}
	else
	    num_millisec = 0;

	if (argc == 10)
	{
	    /* accept_if_condition */
	    strcpy(accept_if, argv[9]);
/* printf("argv[9] = %s\n", accept_if); */
	}


/*************************************************/
/* use ARR_cmd's command cruncher to do the work */
/*************************************************/
	/* The basic idea is to open a file, create a synchronous	*/
	/* data set by adding the list of channels (argv[3]), apply	*/
	/* all special conditions of data retrieval, then access the	*/
	/* data and write it into a tcl list.				*/

    if (pCtx->stop == 0)
    {
	/* command to open archive file */
	strcpy(pCtx->pCmdCtx->line, openFileCommand);
	pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	stat = arrCmdCrunch(pCtx, interp);
	/* If the file doesn't open properly return error */
	if(stat == TCL_ERROR)
	{
	    Tcl_SetResult(interp, "arr_getData: Problem opening archive file", TCL_STATIC);
	    return TCL_ERROR;
	}


	/* Add channels to retrieve data from */
	for(i=0; i< num_pvnames; i++)
	{
	    /* command is of the form "channel pvname" */	
	    sprintf(pCtx->pCmdCtx->line, "channel %s", pvnameList[i]);
/* printf("channel = %s\n", pvnameList[i]); */
	    pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	    nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	    stat = arrCmdCrunch(pCtx, interp);
	    if(stat == TCL_ERROR)
	    {
		Tcl_SetResult(interp, "arr_getData: error adding retrieval channel", TCL_STATIC);
		return TCL_ERROR;
	    }
	}
	free(pvnameList);

	/* Discard Partial Samples if requested */
	if ((argc >= 7) && (discardPartials_flag == 1))
	{
	    strcpy(pCtx->pCmdCtx->line, "discardPartial");
	    pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	    nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	    stat = arrCmdCrunch(pCtx, interp);
	    if(stat == TCL_ERROR)
	    {
		Tcl_SetResult(interp, "arr_getData: error discarding partial samples", TCL_STATIC);
		return TCL_ERROR;
	    }
	}

	/* Means & stdevs if requested */
	if ((argc >= 8) && (mean_stdev_flag == 1))
	{
	    strcpy(pCtx->pCmdCtx->line, "workWith meanAndStdDev");
	    pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	    nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	    stat = arrCmdCrunch(pCtx, interp);
	    if(stat == TCL_ERROR)
	    {
		Tcl_SetResult(interp, "arr_getData: error using means & stdevs", TCL_STATIC);
		return TCL_ERROR;
	    }
	}

	/* Round TimeStamps if requested */
	if ((argc >= 9) && (num_millisec != 0))
	{
	    sprintf(pCtx->pCmdCtx->line, "roundStamps %d", num_millisec);
	    pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	    nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	    stat = arrCmdCrunch(pCtx, interp);
	    if(stat == TCL_ERROR)
	    {
		Tcl_SetResult(interp, "arr_getData: error rounding timestamps", TCL_STATIC);
		return TCL_ERROR;
	    }
	}

	/* "Accept If" condition if requested */
	if (argc == 10)
	{
	    sprintf(pCtx->pCmdCtx->line, "accept if %s", accept_if);
	    pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	    nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	    stat = arrCmdCrunch(pCtx, interp);
	    if(stat == TCL_ERROR)
	    {
		Tcl_SetResult(interp, "arr_getData: error setting \"accept if\" condition", TCL_STATIC);
		return TCL_ERROR;
	    }
	}

	/* command to set starting time */
	sprintf(pCtx->pCmdCtx->line, "position %s", startingTimeStamp);
	pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	stat = arrCmdCrunch(pCtx, interp);
	if(stat == TCL_ERROR)
	{
	    Tcl_SetResult(interp, "arr_getData: error with position command", TCL_STATIC);
	    return TCL_ERROR;
	}

	/* command to get data, "getData", for "num_samples" samples */
	sprintf(pCtx->pCmdCtx->line, "getData %d", num_samples);
	pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	stat = arrCmdCrunch(pCtx, interp);
	if(stat == TCL_ERROR)
	    return TCL_ERROR; /* message set in "sydSampleGet" or "ARR_getData" */

	/* command to quit */
	strcpy(pCtx->pCmdCtx->line, "quit");
	pCtx->pCmdCtx->pLine = pCtx->pCmdCtx->line;
	nextANField(&pCtx->pCmdCtx->pLine, &pCtx->pCmdCtx->pCommand, &pCtx->pCmdCtx->delim);
	stat = arrCmdCrunch(pCtx, interp);
	if(stat == TCL_ERROR)
	{
	    Tcl_SetResult(interp, "arr_getData: error quitting", TCL_STATIC);
	    return TCL_ERROR;
	}
    }

/******************/
/* Wrap things up */
/******************/

    if (pCtx->pSspec != NULL)
    {
	stat = sydClose(pCtx->pSspec);
	assert(stat == S_syd_OK);
    }

    while (pCtx->pCmdCtx->pPrev != NULL)
	cmdCloseContext(&pCtx->pCmdCtx);

    if (pCtx->pCmdCtx->dataOut != stdout)
	fclose(pCtx->pCmdCtx->dataOut);

    /* puts all retrieved data into result string for interp */
    Tcl_DStringResult(interp, &pCtx->returnValue);

    return TCL_OK;
}


/******************************************************************************
* NAME	Arr_TsAddSec
*
* ARGUMENTS
*	TimeStamp TimeSec
*
* DESCRIPTION
*	This subroutine is a "command procedure" in tcl. It follows the tcl
*	protocol in eventually establishing a tcl command, "arr_tsAddSec",
*	which returns a timestamp equal to "TimeStamp" plus the time in
*	seconds given by the second argument (TimeSec) which may be
*	be positive or negative. It uses the code in tsSubr.c  .
*
*	This subroutine "becomes" a tcl command by being declared in the file,
*	/something/arrtk/arrAppInit.c  using Tcl_CreateCommand.
*
* RETURNS
*	TCL_OK or TCL_ERROR
*
********************************************************************************/
#ifdef __STDC__
int
Arr_TsAddSec(
  ClientData	clientData,
  Tcl_Interp	*interp,
  int		argc,
  char		**argv)
#else
int
Arr_TsAddSec(clientData, interp, argc, argv)
  ClientData	clientData;
  Tcl_Interp	*interp;
  int		argc;
  char		**argv;
#endif /*__STDC__*/

{
	TS_STAMP	tsSum, tStamp;
	char		stampText[28];
	double		seconds;

	if(argc != 3)
	{
	    /* send tcl error message */
	    Tcl_SetResult(interp, "arr_tsAddSec: wrong # args", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp, &argv[1]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_tsAddSec: bad timestamp format", TCL_STATIC);
	    return TCL_ERROR;
	}

	tsAddDouble(&tsSum, &tStamp, atof(argv[2]));
	tsStampToText(&tsSum, TS_TEXT_MMDDYY, stampText);
	Tcl_SetResult(interp, stampText, TCL_VOLATILE);
	return TCL_OK;
}


/******************************************************************************
* NAME	Arr_TsCmpStamps
*
* ARGUMENTS
*	TimeStamp1 TimeStamp2
*
* DESCRIPTION
*	This defines a tcl command, "arr_tsCmpStamps", which returns:
*		-1 if TimeStamp1 is earlier than TimeStamp2
*		 0 if TimeStamp1 is the same as  TimeStamp2
*		+1 if TimeStamp1 is later than   TimeStamp2
*
* RETURNS
*	TCL_OK or TCL_ERROR
*
********************************************************************************/
#ifdef __STDC__
int
Arr_TsCmpStamps(
  ClientData	clientData,
  Tcl_Interp	*interp,
  int		argc,
  char		**argv)
#else
int
Arr_TsCmpStamps(clientData, interp, argc, argv)
  ClientData	clientData;
  Tcl_Interp	*interp;
  int		argc;
  char		**argv;
#endif /*__STDC__*/

{
	TS_STAMP	tStamp1, tStamp2;
	long		returnVal;
	char		returnText[3];

	if(argc != 3)
	{
	    /* send tcl error message */
	    Tcl_SetResult(interp, "arr_tsAddSec: wrong # args", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp1, &argv[1]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_tsAddSec: bad timestamp1 format", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp2, &argv[2]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_tsAddSec: bad timestamp2 format", TCL_STATIC);
	    return TCL_ERROR;
	}

	returnVal = (long) tsCmpStamps(&tStamp1, &tStamp2);
	sprintf(returnText, "%ld", returnVal);
	Tcl_SetResult(interp, returnText, TCL_VOLATILE);
	return TCL_OK;
}


/******************************************************************************
* NAME	Arr_DiffAsSec
*
* ARGUMENTS
*	TimeStamp1 TimeStamp2
*
* DESCRIPTION
*	This defines a tcl command, "arr_diffAsSec", which returns:
*	(TimeStamp1 - TimeStamp2) in seconds.
*
* RETURNS
*	TCL_OK or TCL_ERROR
*
********************************************************************************/
#ifdef __STDC__
int
Arr_DiffAsSec(
  ClientData	clientData,
  Tcl_Interp	*interp,
  int		argc,
  char		**argv)
#else
int
Arr_DiffAsSec(clientData, interp, argc, argv)
  ClientData	clientData;
  Tcl_Interp	*interp;
  int		argc;
  char		**argv;
#endif /*__STDC__*/

{
	TS_STAMP	tStamp1, tStamp2;
	char		returnText[20];
	double		difference;

	if(argc != 3)
	{
	    /* send tcl error message */
	    Tcl_SetResult(interp, "arr_diffAsSec: wrong # args", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp1, &argv[1]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_diffAsSec: bad timestamp1 format", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp2, &argv[2]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_diffAsSec: bad timestamp2 format", TCL_STATIC);
	    return TCL_ERROR;
	}

	TsDiffAsDouble(&difference, &tStamp1, &tStamp2);
	sprintf(returnText, "%lg", difference);
	Tcl_SetResult(interp, returnText, TCL_VOLATILE);
	return TCL_OK;
}


/******************************************************************************
* NAME	Arr_DiffAsStamp
*
* ARGUMENTS
*	TimeStamp1 TimeStamp2
*
* DESCRIPTION
*	This defines a tcl command, "arr_diffAsStamp", which returns:
*	(TimeStamp1 - TimeStamp2) in timestamp format. Sign is taken
*	care of.
*
*	Note that all time differences are referenced from 1/1/90, so
*	any time differences greater that 24 hours will be reflected in the
*	date part of the TS relative to this reference date. Also note
*	that a difference of less than 24 hours will omit the date part
*	of the timestamp. 
*
* RETURNS
*	TCL_OK or TCL_ERROR
*
********************************************************************************/
#ifdef __STDC__
int
Arr_DiffAsStamp(
  ClientData	clientData,
  Tcl_Interp	*interp,
  int		argc,
  char		**argv)
#else
int
Arr_DiffAsStamp(clientData, interp, argc, argv)
  ClientData	clientData;
  Tcl_Interp	*interp;
  int		argc;
  char		**argv;
#endif /*__STDC__*/

{
	TS_STAMP	tStamp1, tStamp2, tStampDiff;
	char		returnText[28];

	if(argc != 3)
	{
	    /* send tcl error message */
	    Tcl_SetResult(interp, "arr_diffAsStamp: wrong # args", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp1, &argv[1]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_diffAsStamp: bad timestamp1 format", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (tsTextToStamp(&tStamp2, &argv[2]) != S_ts_OK)
	{
	    Tcl_SetResult(interp, "arr_diffAsStamp: bad timestamp2 format", TCL_STATIC);
	    return TCL_ERROR;
	}

	/* make sure 2nd argument of tsCmpStamps is later that 3rd */
	if (tsCmpStamps(&tStamp1, &tStamp2) >= 0) {
	    TsDiffAsStamp(&tStampDiff, &tStamp1, &tStamp2);
	} else {
	    TsDiffAsStamp(&tStampDiff, &tStamp2, &tStamp1);
	}

	tsStampToText(&tStampDiff, TS_TEXT_MMDDYY, returnText);
	Tcl_SetResult(interp, returnText, TCL_VOLATILE);
	return TCL_OK;
}


/*****************************************************************************
* NAME	arrSigHandler - signal handling and initialization
*
* DESCRIPTION
*	These routines set up for the signals to be caught for arr
*	and handle the signals when they occur.
*
* RETURNS
*	void
*
* BUGS
* o	not all signals are caught
* o	it's not clear how useful it is to catch the signals which originate
*	from the keyboard
*
*******************************************************************************/
#ifdef __STDC__
static void
arrSigHandler(int signo)
#else
static void
arrSigHandler(signo)
  int	signo;
#endif /*__STDC__*/

{
    printf("entered arrSigHandler for signal:%d\n", signo);
    signal(signo, SIG_DFL);
    longjmp(pglArrCtx->sigEnv, 1);
}


/*************************************************************
* NAME	arrCmdCrunch - process a command line
*
* DESCRIPTION
*
* RETURNS
*	0 if command successful
*	1 if command unsuccessful
*
* BUGS
* o	text
*
*************************************************************/
#ifdef __STDC__
static int
arrCmdCrunch(ARR_CTX *pArrCtx, Tcl_Interp *interp)
#else
static int
arrCmdCrunch(pArrCtx, interp)
  ARR_CTX    *pArrCtx;	/* IO pointer to ARR context  */
  Tcl_Interp *interp;	/* pointer to tcl interpreter */
#endif /*__STDC__*/

{
    CX_CMD	*pCmdCtx=pArrCtx->pCmdCtx;
    SYD_SPEC	*pSspec;
    long	stat;

    pSspec = pArrCtx->pSspec;

    if (strcmp(P_CMD,				"assert0") == 0) {
/*-------------------------------------------------------------------
*	under SunOS, assertAlways() generates SIGABRT
*--------------------------------------------------------------------*/
	assertAlways(0);
    }
    else if (strcmp(P_CMD,			"quit") == 0) {
	pArrCtx->stop = 1;
	return TCL_OK;
    }
    else if (strcmp(P_CMD,			"trap") == 0) {
	int	*j;
	j = (int *)(-1);
	j = (int *)(*j);
	return TCL_OK;
    }
    else if (strcmp(P_CMD,			"dataOut") == 0) {
	if (pCmdCtx->dataOut != stdout)
	    fclose (pCmdCtx->dataOut);
	if (nextNonSpaceField(&P_LINE, &P_FLD, &DELIM) < 1)
	    pCmdCtx->dataOut = stdout;
	else {
	    pCmdCtx->dataOut = fopen(pCmdCtx->pField, "a");
	    if (pCmdCtx->dataOut == NULL) {
		printf("couldn't open %s\n", pCmdCtx->pField);
		pCmdCtx->dataOut = stdout;
	    }
	}
	return TCL_OK;
    }
    else if (strcmp(P_CMD,			"source") == 0) {
	cmdSource(&pArrCtx->pCmdCtx);
	return TCL_OK;
    }
    else if (strcmp(P_CMD,			"byChannel") == 0) {
	if (pArrCtx->pSspec != NULL) {
	    stat = sydClose(pArrCtx->pSspec);
	    assertAlways(stat == S_syd_OK);
	    pArrCtx->pSspec = NULL;
	}
	
	if (nextNonSpaceField(&P_LINE, &P_FLD, &DELIM)>1) {
	    if (strlen(P_FLD) >= 80) {
		printf("file name (or path) too long\n");
		return TCL_ERROR;
	    }
	    (void)strcpy(pArrCtx->bychanFile, P_FLD);
	}
	else
	    (void)strcpy(pArrCtx->bychanFile, "arChan/arChanData.bin");
	stat = sydOpenCF(&pArrCtx->pSspec, pArrCtx->bychanFile);
	if (stat != S_syd_OK) {
	    printf("initialization failed\n");
	    return TCL_ERROR;
	}
	return TCL_OK;
    }
    else if (strcmp(P_CMD,			"sampleSet") == 0) {
	if (pArrCtx->pSspec != NULL) {
	    stat = sydClose(pArrCtx->pSspec);
	    assertAlways(stat == S_syd_OK);
	    pArrCtx->pSspec = NULL;
	}
	if (nextNonSpaceField(&P_LINE, &P_FLD, &DELIM) <= 1) {
	    printf("file name (or path) must be specified\n");
	    return TCL_ERROR;
	}
	if (strlen(P_FLD) >= 80) {
	    printf("file name (or path) too long\n");
	    return TCL_ERROR;
	}
	(void)strcpy(pArrCtx->sampleFile, P_FLD);
	stat = sydOpenSSF(&pArrCtx->pSspec, pArrCtx->sampleFile);
	if (stat != S_syd_OK) {
	    printf("initialization failed\n");
	    return TCL_ERROR;
	}
	return TCL_OK;
    }
    if (pSspec == NULL && strcmp(P_CMD, "help") != 0) {
	printf( "no file selected yet; use `byChannel' or `sampleSet'\n");
	return TCL_ERROR;
    }
    if (strcmp(P_CMD,				"channel") == 0)
	ARR_channel(pArrCtx);
    else if (strcmp(P_CMD,			"delete") == 0) {
	if (pSspec->chanCount <= 0) goto noChanErr;
	ARR_delete(pArrCtx);
    }
    else if (strcmp(P_CMD,			"fileInfo") == 0 ||
	strcmp(P_CMD, "fileinfo") == 0) {
	(void)sydFileInfo(pSspec, stdout);
    }

/**************************/
/* commands deleted by CT */
/**************************/

    else if (strcmp(P_CMD,			"position") == 0) {
	if (pSspec->chanCount <= 0) goto noChanErr;
	ARR_position(pArrCtx);
    }
    else if (strcmp(P_CMD,			"getData") == 0) {
	if (pSspec->chanCount <= 0) goto noChanErr;
	if(ARR_getData(pArrCtx, interp) == TCL_ERROR)
		return TCL_ERROR;
    }
    else if (strcmp(P_CMD,			"resetFilled") == 0) {
	double	dead;
	if (pSspec->chanCount <= 0) goto noChanErr;
	if (nextFltFieldAsDbl(&P_LINE, &dead, &DELIM) <= 1) {
	    printf("deadtime interval must be specified\n");
	    return TCL_ERROR;
	}
	if (dead <= 0.) {
	    printf("deadtime must be > 0.\n");
	    return TCL_ERROR;
	}
	pArrCtx->deadRstSec = dead;
	pArrCtx->deadRst = 1;
    }
    else if (strcmp(P_CMD,			"accept") == 0) {
	sydTestClose(pSspec);
	strcpy(pArrCtx->acceptText, P_LINE);
	if (strncmp(pArrCtx->acceptText, "all", 3) == 0) {
	    pArrCtx->accept = 0;
	    sydTestClose(pSspec);
	}
	else {
	    stat = sydTestAddFromText(pSspec, pArrCtx->acceptText);
	    if (stat != S_syd_OK) {
		(void)printf("illegal `accept if' condition\n");
		sydTestClose(pSspec);
		pArrCtx->accept = 0;
	    }
	    else
		pArrCtx->accept = 1;
	}
    }
    else if (strcmp(P_CMD,			"discardPartial") == 0) {
	pArrCtx->discardPartial = 1;
    }
    else if (strcmp(P_CMD,			"rewind") == 0) {
	if (pSspec->chanCount <= 0) goto noChanErr;
	sydPosition(pSspec, NULL);
    }
    else if (strcmp(P_CMD,			"roundStamps") == 0) {
	long	round;
	if (pSspec->chanCount <= 0) goto noChanErr;
	if (nextIntFieldAsLong(&P_LINE, &round, &DELIM) <= 1) {
	    printf("rounding interval must be specified\n");
	    return TCL_ERROR;
	}
	if (round < 0 || round > 2000) {
	    printf("0 < rounding interval <= 2000 violated\n");
	    return TCL_ERROR;
	}
	pSspec->roundNsec = round * 1000000;
    }
    else if (strcmp(P_CMD,			"workWith") == 0) {
	if (nextNonSpaceField(&P_LINE, &P_FLD, &DELIM) < 1) {
	    printf("`workWith' must have a parameter\n");
	    return TCL_ERROR;
	}
	else if (strcmp(P_FLD, "dataVal") == 0)
	    pArrCtx->useStats = pArrCtx->useMeans = 0;
	else if (strcmp(P_FLD, "mean") == 0)
	    pArrCtx->useStats = 0, pArrCtx->useMeans = 1;
	else if (strcmp(P_FLD, "meanAndStdDev") == 0)
	    pArrCtx->useStats = 1, pArrCtx->useMeans = 0;
	else {
	    printf("illegal `workWith' parameter\n");
	    return TCL_ERROR;
	}
    }
    else {
/*----------------------------------------------------------------------------
* help (or illegal command)
*----------------------------------------------------------------------------*/
	(void)nextNonSpaceField(&P_LINE, &P_FLD, &DELIM);
	helpIllegalCommand(stdout, &pCmdCtx->helpList, P_CMD, P_FLD);
    }

    return TCL_OK;
noChanErr:
    printf("no channels selected\n");
}


/**************************************************************
* NAME	arrInitAtStartup - initialization for ARR
*
* DESCRIPTION
*	Perform several initialization duties:
*	o   initialize the channel data retrieval set
*	o   initialize the command context block
*	o   initialize the help information
*
* RETURNS
*	OK
*
* BUGS
* o	text
*****************************************************************/
#ifdef __STDC__
static long
arrInitAtStartup(ARR_CTX *pArrCtx)
#else
static long
arrInitAtStartup(pArrCtx)
  ARR_CTX	*pArrCtx;
#endif /*__STDC__*/

{
    long	stat;
    CX_CMD	*pCmdCtx = &pArrCtx->cmdCtx;

    pArrCtx->pCmdCtx = pCmdCtx;
    pArrCtx->stop = 0;
    pArrCtx->accept = 0;
    pArrCtx->deadRst = 0;
    pArrCtx->discardPartial = 0;
    pArrCtx->useStats = pArrCtx->useMeans = 0;

    pArrCtx->sampleFile[0] = '\0';
    pArrCtx->pSspec = NULL;

    cmdInitContext(pCmdCtx, " ARR_cmd: ");

    Tcl_DStringInit(&pArrCtx->returnValue);
    pArrCtx->numberSamples = 0;

    return OK;
}


/*------------------------------------------------------------
* channel
*-------------------------------------------------------------*/
#ifdef __STDC__
static void
ARR_channel(ARR_CTX *pArrCtx)
#else
static void
ARR_channel(pArrCtx)
  ARR_CTX *pArrCtx;	/* IO pointer to ARR context */
#endif /*__STDC__*/

{
    CX_CMD	*pCmdCtx=pArrCtx->pCmdCtx;
    SYD_SPEC	*pSspec=pArrCtx->pSspec;
    long	stat;
    int		syncType;

    sydSampleSetFree(pSspec);

    if ((FLD_LEN = nextChanNameField(&P_LINE, &P_FLD, &DELIM)) <= 1)
	(void)printf("you must specify one or more channels\n");
    while (FLD_LEN > 1) {
	syncType = SYD_SY_NONF;
	if (strcmp(P_FLD, "F") == 0) {
	    syncType = SYD_SY_FILLED;
	    FLD_LEN = nextChanNameField(&P_LINE, &P_FLD, &DELIM);
	}
	if (P_FLD[0] == '#')
	    break;
	else if (sydChanFind(pSspec, P_FLD) != NULL)
	    (void)printf("%s already selected\n", P_FLD);
	else {
	    stat = sydChanOpen(pSspec, NULL, P_FLD, syncType, NULL,0);
	    if (stat != S_syd_OK)
		(void)printf("couldn't open %s \n", P_FLD);
	    else
		pSspec->chanCount++;
	}
	FLD_LEN = nextChanNameField(&P_LINE, &P_FLD, &DELIM);
    }
    return;
}


/*-------------------------------------------------------------
* delete
*--------------------------------------------------------------*/
#ifdef __STDC__
static void
ARR_delete(ARR_CTX *pArrCtx)
#else
static void
ARR_delete(pArrCtx)
  ARR_CTX *pArrCtx;	/* IO pointer to ARR context */
#endif /*__STDC__*/

{
    CX_CMD	*pCmdCtx=pArrCtx->pCmdCtx;
    SYD_SPEC	*pSspec=pArrCtx->pSspec;
    SYD_CHAN	*pSChan, *pSChanNext;

    FLD_LEN = nextChanNameField(&P_LINE, &P_FLD, &DELIM);
    if (FLD_LEN <= 1) {
	(void)printf(
		"you must specify one or more channels or else specify all\n");
	return;
    }
    if (strcmp(P_FLD, "all") == 0) {
	pSChan = pSspec->pChanHead;
	while (pSChan != NULL) {
	    pSChanNext = pSChan->pNext;
	    if (sydChanClose(pSspec, pSChan) != S_syd_OK) {
		(void)printf("error deleting %s\n", P_FLD);
	    }
	    pSChan = pSChanNext;
	}
	return;
    }
    while (FLD_LEN > 1) {
	if ((pSChan = sydChanFind(pSspec, P_FLD)) == NULL)
	    (void)printf("%s not selected\n", P_FLD);
	else {
	    if (sydChanClose(pSspec, pSChan) != S_syd_OK) {
		(void)printf("error deleting %s\n", P_FLD);
	    }
	}
	FLD_LEN = nextChanNameField(&P_LINE, &P_FLD, &DELIM);
    }
}


/*--------------------------------------------------------
* position
*---------------------------------------------------------*/
#ifdef __STDC__
static void
ARR_position(ARR_CTX *pArrCtx)
#else
static void
ARR_position(pArrCtx)
  ARR_CTX *pArrCtx;	/* IO pointer to ARR context */
#endif /*__STDC__*/

{
    CX_CMD	*pCmdCtx=pArrCtx->pCmdCtx;
    SYD_SPEC	*pSspec=pArrCtx->pSspec;
    TS_STAMP	posStamp;
    double	deltaTime;
    int		stat;
    SYD_CHAN	*pSChan;

    (void)nextNonSpace(&P_LINE);
    if (tsTextToStamp(&posStamp, &P_LINE) == S_ts_OK)
	sydPosition(pSspec, &posStamp);
    else
	(void)printf("error in time specification\n");
}


/*---------------------------------------------------------
* sample
*----------------------------------------------------------*/
#ifdef __STDC__
static int
ARR_sample(ARR_CTX *pArrCtx, int reqCount)
#else
static int
ARR_sample(pArrCtx, reqCount)
  ARR_CTX *pArrCtx;	/* IO pointer to ARR context */
  int	  reqCount;
#endif /*__STDC__*/

{
    CX_CMD	*pCmdCtx=pArrCtx->pCmdCtx;
    SYD_SPEC	*pSspec=pArrCtx->pSspec;
    long	stat;
    int 	done=0;
    double	diff;
    
/*-----------------------------------------------------------------------------
*	start time fresh for all channels
*----------------------------------------------------------------------------*/
    sydInputResetSampled(pSspec);
/*----------------------------------------------------------------------------
*	reserve memory for the various sample storage arrays if there is
*	no storage or its size has increased
*----------------------------------------------------------------------------*/
    if (sydSampleSetAlloc(pSspec, reqCount) != S_syd_OK)
	return ERROR;
/*----------------------------------------------------------------------------
*	now, gather the data  (note that if `accept if' was specified, it
*	was applied to pSspec when the command was encountered)
*----------------------------------------------------------------------------*/
    while (!done) {
	stat = sydInputGet(pSspec, NULL);
	if (pArrCtx->deadRst == 1) {
	    if (pArrCtx->deadRefTs.secPastEpoch == 0)
		pArrCtx->deadRefTs = SydInputTs(pSspec);
	    else {
		TsDiffAsDouble(&diff,&SydInputTs(pSspec),&pArrCtx->deadRefTs);
		if (diff > pArrCtx->deadRstSec) {
		    sydInputResetSampled(pSspec);
		    stat = sydInputGet(pSspec, NULL);
		}
		pArrCtx->deadRefTs = SydInputTs(pSspec);
	    }
	}
	if (stat == S_syd_OK || stat == S_syd_partial) {
	    sydInputStoreInSet(pSspec, pArrCtx->discardPartial);
	    if (pSspec->sampleCount >= pSspec->reqCount)
		done = 1;
	}
	else
	    done = 1;
    }
}


/*----------------------------------------------------------------------------
*
* NAME
*	ARR_getData
*
* DESCRIPTION
*
*	This routine obtains the data samples in a sample set.
*	If the `useStats' flag is set in the sync set spec, then means
*	and standard deviations are obtained instead of the actual data
*	values. If the `useMeans' and `useStats' flags are set, then
*	only the means are obtained.
*
* RETURNS
*	TCL_OK
* 
*----------------------------------------------------------------------------*/
#ifdef __STDC__
static int
ARR_getData(ARR_CTX *pArrCtx, Tcl_Interp *interp)
#else
static int
ARR_getData(pArrCtx, interp)
  ARR_CTX    *pArrCtx;	/* IO pointer to ARR context  */
  Tcl_Interp *interp;	/* pointer to tcl interpreter */
#endif /*__STDC__*/

{
    CX_CMD	*pCmdCtx=pArrCtx->pCmdCtx;
    SYD_SPEC	*pSspec=pArrCtx->pSspec;/* pointer to synchronous set spec */
    int		reqCount;

    assert(pSspec != NULL);

    if (nextIntFieldAsInt(&P_LINE, &reqCount, &DELIM) > 1) {
	if (reqCount > 10000) {
	    printf("sample count must be <= 10000\n");
	    /* send tcl error message */
	    Tcl_SetResult(interp, "sample count must be <= 10000", TCL_STATIC);
	    return TCL_ERROR;
	}
	ARR_sample(pArrCtx, reqCount);
    }

    if (pSspec->sampleCount <= 0)
    {
	printf("no data to get\n");
	/* send tcl error message */
	Tcl_SetResult(interp, "no data to get", TCL_STATIC);
	return TCL_ERROR;
    }

    if (pArrCtx->useStats == 0 && pArrCtx->useMeans == 0)
	sydSetAttr(pSspec, SYD_ATTR_USE_STATS, 0, NULL);
    else {
	if (pArrCtx->useStats)
	    sydSetAttr(pSspec, SYD_ATTR_USE_STATS, 1, NULL);
	else
	    sydSetAttr(pSspec, SYD_ATTR_USE_MEANS, 1, NULL);
	sydSampleSetStats(pSspec);
    }

    if (pSspec->useStats)
    {
	if (ARR_getStatData(pArrCtx, interp) != TCL_OK)
	    return TCL_ERROR;
    }
    else
    {
	if (ARR_getRegularData(pArrCtx, interp) != TCL_OK)
	    return TCL_ERROR;
    }

    return TCL_OK;
}


/*-----------------------------------------------------------------------------
*
* NAME
*	ARR_getStatData
*
* DESCRIPTION
*
* 	This routine obtains the statistical data in a sample set.
*	If the `useStats' flag is set in the sync set spec, then means and
*	standard deviations are obtained instead of the actual data values.
*
*	This routine writes all the data returned by the tk command
*	"arr_getData", into the global tcl dynamic string "returnData".
*	This is necessary as otherwise data needs to travel too far up
*	the calling sequence.
*
*
* RETURNS
*	TCL_OK
* 
*----------------------------------------------------------------------------*/
#ifdef __STDC__
static int
ARR_getStatData(ARR_CTX *pArrCtx, Tcl_Interp *interp)
#else
static int
ARR_getStatData(pArrCtx, interp)
  ARR_CTX	*pArrCtx;	/* IO pointer to ARR context	*/
  Tcl_Interp	*interp;	/* pointer to tcl interpreter	*/
#endif /*__STDC__*/

{
    SYD_SPEC	*pSspec = pArrCtx->pSspec;  /* pointer to sync set spec */
    SYD_CHAN	*pSChan;	/* pointer to channel in sync set	*/
    SYD_CHAN	*pNext;

    Tcl_DString pvNames,	/*					*/
		masterTimes,	/* tcl dynamic strings are used to hold	*/
		masterTstamps,	/* data lists which are combined into a	*/
		statCounts,	/* complete list (returnData) returned	*/
		units,		/* as part of the arrGetData command.	*/
		values,
		valuesSub,
		times,
		tStamps,
		mins,
		maxs,
		stdevs,
		stdevsSub;
		
    int		index,		/* for counting				*/
		prec,		/* precision for mean & stdev display	*/
		sampBegin,	/* sample # of beginning data		*/
		sampEnd;	/* sample # of ending data		*/
    chtype	type;		/* type of value */
    long	samp,		/* sample number in synchronous set	*/
		totalNumSamp;	/* total # of samps (user's) collected	*/
    char	text[28];	/* used to hold strings temporarily	*/
    double	minVal, maxVal,	/* min & max of a pv in a sync set	*/
		mean, stdev;	/* mean value for a channel in a sample	*/

    /* Initialize all Tcl dynamic strings */
    Tcl_DStringInit(&pvNames);
    Tcl_DStringInit(&masterTimes);	Tcl_DStringInit(&masterTstamps);
    Tcl_DStringInit(&statCounts);	Tcl_DStringInit(&units);
    Tcl_DStringInit(&values);		Tcl_DStringInit(&valuesSub);
    Tcl_DStringInit(&times);		Tcl_DStringInit(&tStamps);
    Tcl_DStringInit(&mins);		Tcl_DStringInit(&maxs);
    Tcl_DStringInit(&stdevs);		Tcl_DStringInit(&stdevsSub);

    /*******************************************************************/
    /* Scanning through samples, get times, timestamps, and statcounts */
    /*******************************************************************/
    /* This is the "normal" data retrieval mode with a starting TS */
    /* and the #_of_samples forward in time that are desired.	   */
    if (strcmp(pArrCtx->retrievalCode, "NUMBER+") == 0)
    {
      for (samp = 0; samp < pSspec->statCount; samp++)
      {
	/* This becomes "masterTime_ts" in arr - list of all timestamps. */
	tsStampToText(&pSspec->pStatTimeStamp[samp], TS_TEXT_MMDDYY, text);
	Tcl_DStringAppendElement(&masterTstamps, text);

	/* This becomes "masterTime_sec" in arr - list of all times (sec) */
	sprintf(text, "%lg", pSspec->pStatDeltaSec[samp]);
	Tcl_DStringAppendElement(&masterTimes, text);

	/* This becomes "statCount" - list of #_samples in each mean/stdev calc. */
	sprintf(text, "%ld", pSspec->pStatPopCount[samp]);
	Tcl_DStringAppendElement(&statCounts, text);
      }
      sampBegin = 0;
      sampEnd   = pSspec->statCount - 1;
    }

    /* This is the retrieval mode with a starting & an ending TS */
    else if (strcmp(pArrCtx->retrievalCode, "TIMESTAMP") == 0)
    {
      index = 0;

      for (samp = 0; samp < pSspec->statCount; samp++)
      {
	/* For data obtained by specifing beginning & ending TS's,	*/
	/* filter out the unwanted timestamps. Limitations of the 	*/
	/* existing archive software force us to search by specifing a	*/
	/* starting TS and the # of samples desired. Since we want an	*/
	/* ending TS, we simply set the # of samples to its max value	*/
	/* (10,000) and ignore data after the ending TS.		*/

	/* if TS is earlier than desired starting TS, go to next TS */
	if (tsCmpStamps(&pSspec->pStatTimeStamp[samp], &pArrCtx->startingTS) == -1)
	    continue;
	/* if TS is later than desired ending TS, quit */		
	else if (tsCmpStamps(&pSspec->pStatTimeStamp[samp], &pArrCtx->endingTS) == 1)
	    break;
	else
	{
	    /* keep track of first data point sample # */
	    if (index == 0) sampBegin = samp;
	    /* keep track of last data point sample # */
	    sampEnd = samp;

	    index++;

	    tsStampToText(&pSspec->pStatTimeStamp[samp], TS_TEXT_MMDDYY, text);
	    Tcl_DStringAppendElement(&masterTstamps, text);

	    sprintf(text, "%lg", pSspec->pStatDeltaSec[samp]);
	    Tcl_DStringAppendElement(&masterTimes, text);

	    sprintf(text, "%ld", pSspec->pStatPopCount[samp]);
	    Tcl_DStringAppendElement(&statCounts, text);
	}
      }
    }

    /* This is the retrieval mode (NUMBER-) with a starting TS	*/
    /* and the #_of_samples backward in time that is desired.	*/

    /* NOTE! : The #_of_samples specified by user is NOT the	*/
    /* same as "samp" below. Each statistical "samp" contains	*/
    /* many #_of_samples used to calculate the mean & stdev.	*/

    else
    {
      index = 0;
      totalNumSamp = 0;

      /* Samples are scanned in reverse to easily get beginning & ending samp #'s */
      for (samp = pSspec->statCount - 1; samp >= 0; samp--)
      {
	/* if TS is later than desired "starting" TS, go to next TS */		
	if (tsCmpStamps(&pSspec->pStatTimeStamp[samp], &pArrCtx->startingTS) == 1) 
	    continue;
	/* If desired #_of_samples have been collected, quit.	*/
	/* Note that this can only be done approximately.	*/
	else if (totalNumSamp >= pArrCtx->numberSamples)
	    break;
	else
	{
	    /* last data point sample # */
	    if (index == 0) sampEnd = samp;
	    /* first data point sample # */
	    sampBegin = samp;
	    /* total # of user-specified samples collected so far */
	    totalNumSamp = totalNumSamp + pSspec->pStatPopCount[samp];

	    index++;
	}
      }

      /* Rescan forwards to get timestamps in the correct chronological order */
      for (samp = sampBegin; samp <= sampEnd; samp++)
      {
	    tsStampToText(&pSspec->pStatTimeStamp[samp], TS_TEXT_MMDDYY, text);
	    Tcl_DStringAppendElement(&masterTstamps, text);

	    sprintf(text, "%lg", pSspec->pStatDeltaSec[samp]);
	    Tcl_DStringAppendElement(&masterTimes, text);

	    sprintf(text, "%ld", pSspec->pStatPopCount[samp]);
	    Tcl_DStringAppendElement(&statCounts, text);
      }
    }

    /* scan through channels to get means (values) and stdevs */
    for (pSChan=pSspec->pChanHead; pSChan!=NULL; pSChan=pSChan->pNext)
    {
	/* re-initialize strings */
	Tcl_DStringFree(&valuesSub);
	Tcl_DStringFree(&stdevsSub);

	/* only look at numerical data ... */
	type = pSChan->dbrType;
	if ((type != DBR_TIME_STRING)&&(type != DBR_TIME_ENUM)&&(pSChan->dataChan))
	{
	    /* This string becomes supplies all pvnames */
	    Tcl_DStringAppendElement(&pvNames, pSChan->name);

	    /* This string becomes "units" - an arr list */
	    Tcl_DStringAppendElement(&units, pSChan->EGU);

	    /* reset for each channel */
	    maxVal = -1.E99; minVal = 1.E99;

	    /* go through all samples */
	    for (samp = sampBegin; samp <= sampEnd; samp++)
	    {
		/* Find max & min for each channel.   */
		/* For statistics, this program needs */
		/* to find its own max & min values   */
		/* since there are none supplied in   */
		/* the archiver's structures.         */
		mean  = pSChan->pStats[samp].mean;
		stdev = pSChan->pStats[samp].stdDev;

		if (minVal > mean)
		    minVal = mean;
		if (maxVal < mean)
		    maxVal = mean;

		/* This becomes part of "value(pvname)" - an arr list */
		sprintf(text, "%lg", mean);
		Tcl_DStringAppendElement(&valuesSub, text);

		if (pSspec->useMeans == 0)
		{
		    /* This becomes part of "stdev(pvname)" - an arr list */
		    sprintf(text, "%.3lg", stdev);
		    Tcl_DStringAppendElement(&stdevsSub, text);
		}
	    }

	    /* This string becomes "minimum" - an arr list */
	    sprintf(text, "%lg", minVal);
	    Tcl_DStringAppendElement(&mins, text);

	    /* This string becomes "maximum" - an arr list */
	    sprintf(text, "%lg", maxVal);
	    Tcl_DStringAppendElement(&maxs, text);
	}

	/* Place temporary list into complete list (as an element). */
	/* In the case of statical data, the master time & tstamp   */
	/* lists are the same as the lists for each channel.	    */
	Tcl_DStringAppendElement(&values,  Tcl_DStringValue(&valuesSub));
	Tcl_DStringAppendElement(&stdevs,  Tcl_DStringValue(&stdevsSub));
	Tcl_DStringAppendElement(&times,   Tcl_DStringValue(&masterTimes));
	Tcl_DStringAppendElement(&tStamps, Tcl_DStringValue(&masterTstamps));
    }

    /* For statistics, this program needs to find its own max & min */
    /* values since there are none supplied in the structure.       */

    /* put all lists together in a list of lists (of lists) */
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&pvNames));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&masterTstamps));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&masterTimes));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&tStamps));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&times));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&values));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&units));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&mins));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&maxs));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&stdevs));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&statCounts));

    /* Free up memory */
    Tcl_DStringFree(&pvNames);
    Tcl_DStringFree(&masterTstamps);
    Tcl_DStringFree(&masterTimes);
    Tcl_DStringFree(&tStamps);
    Tcl_DStringFree(&times);
    Tcl_DStringFree(&values);
    Tcl_DStringFree(&units);
    Tcl_DStringFree(&mins);
    Tcl_DStringFree(&maxs);
    Tcl_DStringFree(&stdevs);
    Tcl_DStringFree(&statCounts);
    Tcl_DStringFree(&valuesSub);
    Tcl_DStringFree(&stdevsSub);

    /* return value from this routine */
    return TCL_OK;
}


/*-------------------------------------------------------------------------
* NAME	ARR_getRegularData
*
* DESCRIPTION
*	Obtains the data for all channels in a particular sample. 
*	If chan is an array, get first value only. Does NOT handle
*	statistics.
*
* RETURNS
*	TCL_OK
*
*-------------------------------------------------------------------------*/
#ifdef __STDC__
static int
ARR_getRegularData(ARR_CTX *pArrCtx, Tcl_Interp *interp)
#else
static int
ARR_getRegularData(pArrCtx, interp)
  ARR_CTX	*pArrCtx;	/* IO pointer to ARR context	*/
  Tcl_Interp	*interp;	/* pointer to tcl interpreter	*/
#endif /*__STDC__*/

{
    SYD_SPEC	*pSspec = pArrCtx->pSspec;  /* pointer to sync set spec */
    SYD_CHAN	*pSChan,	/* pointer to channel in synchronous set*/
		*pNext;
    char	text[28],	/* used to hold strings temporarily	*/
		stampText[28],	/* used to hold timestamp text 		*/
		data[20],	/* text version of data value retrieved	*/
		datastatL[2],	/* data Status code Left		*/
		datastatR[2],	/* data Status code Right		*/
		*special,	/* problem indication			*/
		*strVal;	/* data	- type char			*/

    int		sub = 0,	/* subscript for array channels	-	*/
				/* look at first value only		*/
		index = 0,	/* for counting				*/
		sampBegin,	/* sample # of beginning data		*/
		sampEnd,	/* sample # of ending data		*/
		skipDataCollection = 0,	/* if = 1 don't collect data	*/
		myType;		/* 0,1,2 for lng, dbl, str		*/
    long	i,
		lngVal,		/* data - type long			*/
		samp;		/* sample number			*/
    double	dblVal,		/* data - type double			*/
		minVal, maxVal;	/* min & max of a pv in a sync set	*/

    chtype	type;		/* type of value
			*/
    Tcl_DString masterTimes,	/* tcl dynamics strings, used to hold	*/
		masterTstamps,	/* data lists which are combined into a	*/
		pvNames,	/* complete list (returnData) returned	*/
		units,		/* as part of the arr_getData command.	*/
		values,
		valuesSub,
		times,
		timesSub,
		tStamps,
		tStampsSub,
		mins,
		maxs,
		types,
		dStat1,
		dStat1Sub,
		dStat2,
		dStat2Sub;

    /* Initialize all Tcl dynamic strings */
    Tcl_DStringInit(&masterTimes);	Tcl_DStringInit(&masterTstamps);
    Tcl_DStringInit(&pvNames);		Tcl_DStringInit(&units);
    Tcl_DStringInit(&values);		Tcl_DStringInit(&valuesSub);
    Tcl_DStringInit(&times);		Tcl_DStringInit(&timesSub);
    Tcl_DStringInit(&tStamps);		Tcl_DStringInit(&tStampsSub);
    Tcl_DStringInit(&mins);		Tcl_DStringInit(&maxs);
    Tcl_DStringInit(&dStat1);		Tcl_DStringInit(&dStat1Sub);
    Tcl_DStringInit(&dStat2);		Tcl_DStringInit(&dStat2Sub);
    Tcl_DStringInit(&types);

    /* This is the "normal" data retrieval mode with a starting TS */
    /* and the #_of_samples forward in time that are desired.	   */
    if (strcmp(pArrCtx->retrievalCode, "NUMBER+") == 0)
    {
      for (samp = pSspec->restrictFirstData; samp <= pSspec->restrictLastData; samp++)
      {
	    /* This becomes "masterTime_ts" in arr - list of all timestamps. */
	    tsStampToText(&pSspec->pTimeStamp[samp], TS_TEXT_MMDDYY, text);
	    Tcl_DStringAppendElement(&masterTstamps, text);

	    /* This becomes "masterTime_sec" in arr - list of all times (sec) */
	    /* Seconds are converted to text with 12 significant digits */
	    sprintf(text, "%lg", pSspec->pDeltaSec[samp]);
	    Tcl_DStringAppendElement(&masterTimes, text);
      }
      sampBegin = pSspec->restrictFirstData;
      sampEnd   = pSspec->restrictLastData;
    }

    /* This is the retrieval mode with a starting & an ending TS */
    else if (strcmp(pArrCtx->retrievalCode, "TIMESTAMP") == 0)
    {
      index = 0;

      for (samp = pSspec->restrictFirstData; samp <= pSspec->restrictLastData; samp++)
      {
	/* For data obtained by specifing beginning & ending TS's,	*/
	/* filter out the unwanted timestamps. Limitations of the 	*/
	/* existing archive software force us to search by specifing a	*/
	/* starting TS and the # of samples desired. Since we want an	*/
	/* ending TS, we simply set the # of samples to its max value	*/
	/* (10,000) and ignore data after the ending TS.		*/

	/* if TS is earlier than desired starting TS, go to next TS */
	if (tsCmpStamps(&pSspec->pTimeStamp[samp], &pArrCtx->startingTS) == -1)
	    continue;
	/* if TS is later than desired ending TS, quit */		
	else if (tsCmpStamps(&pSspec->pTimeStamp[samp], &pArrCtx->endingTS) == 1)
	    break;
	else
	{
	    /* keep track of first data point sample # */
	    if (index == 0) sampBegin = samp;
	    /* keep track of last data point sample # */
	    sampEnd = samp;

	    index++;

	    tsStampToText(&pSspec->pTimeStamp[samp], TS_TEXT_MMDDYY, text);
	    Tcl_DStringAppendElement(&masterTstamps, text);

	    sprintf(text, "%lg", pSspec->pDeltaSec[samp]);
	    Tcl_DStringAppendElement(&masterTimes, text);
	}
      }

      /* if no data between the 2 specified TS's, skip collecting data */
      if (index == 0) skipDataCollection = 1;
    }

    /* This is the retrieval mode (NUMBER-) with a starting TS	*/
    /* and the #_of_samples backward in time that is desired.	*/
    else
    {
      index = 0;

      /* Samples are scanned in reverse to easily get beginning & ending samp #'s */
      for (samp = pSspec->restrictLastData; samp >= pSspec->restrictFirstData; samp--)
      {
	/* if TS is later than desired "starting" TS, go to next TS */		
	if (tsCmpStamps(&pSspec->pTimeStamp[samp], &pArrCtx->startingTS) == 1) 
	    continue;
	/* if the desired # of samples have been collected, quit */
	else if (index >= pArrCtx->numberSamples)
	    break;
	else
	{
	    /* keep track of last data point sample # */
	    if (index == 0) sampEnd = samp;
	    /* keep track of first data point sample # */
	    sampBegin = samp;

	    index++;
	}
      }

      /* if no data available, skip collecting data */
      if (index == 0) skipDataCollection = 1;

      /* Rescan forwards to get timestamps in the correct chronological order */
      if (skipDataCollection != 1)
      {
	for (samp = sampBegin; samp <= sampEnd; samp++)
	{
	    tsStampToText(&pSspec->pTimeStamp[samp], TS_TEXT_MMDDYY, text);
	    Tcl_DStringAppendElement(&masterTstamps, text);

	    sprintf(text, "%lg", pSspec->pDeltaSec[samp]);
	    Tcl_DStringAppendElement(&masterTimes, text);
	}
      }
    }

    /* Store units, pvname & type information */
    for (pSChan=pSspec->pChanHead; pSChan!=NULL; pSChan=pSChan->pNext)
    {
	/* This string stores pv names in a list */
	Tcl_DStringAppendElement(&pvNames, pSChan->name);

	/* This string becomes "units" - an arr list */
	Tcl_DStringAppendElement(&units, pSChan->EGU);

	/* This string becomes "type" - an arr list */
	if      (pSChan->dbrType == DBR_TIME_STRING) {strcpy(text, "string");}
	else if (pSChan->dbrType == DBR_TIME_FLOAT)  {strcpy(text, "float") ;}
	else if (pSChan->dbrType == DBR_TIME_SHORT)  {strcpy(text, "short") ;}
	else if (pSChan->dbrType == DBR_TIME_DOUBLE) {strcpy(text, "double");}
	else if (pSChan->dbrType == DBR_TIME_LONG)   {strcpy(text, "long")  ;}
	else if (pSChan->dbrType == DBR_TIME_CHAR)   {strcpy(text, "long")  ;}
	else if (pSChan->dbrType == DBR_TIME_ENUM)   {strcpy(text, "string");}
	else {}
	Tcl_DStringAppendElement(&types, text);

	/* switch (pSChan->sync)
	{
	    case SYD_SY_NONF   :	printf("not filled pv\n"); break;
	    case SYD_SY_FILLED :	printf("filled pv\n"); break;
	    case SYD_SY_PAST   :
	    case SYD_SY_UNDEF  :	printf("sync undefined\n"); break;
	}
	*/
    }

    /* scan through all channels */
    for (pSChan=pSspec->pChanHead; pSChan!=NULL; pSChan=pSChan->pNext)
    {
      /* if there is no data, exit loop, set return string, and return */
      if (skipDataCollection == 1)
	break;

      /* reset max & min for each channel */
      maxVal = -1.E99; minVal = 1.E99;

      /* this channel's data type */
      type = pSChan->dbrType;

      /* re-initialize strings for each channel */
      Tcl_DStringFree(&valuesSub);
      Tcl_DStringFree(&timesSub);
      Tcl_DStringFree(&tStampsSub);
      Tcl_DStringFree(&dStat1Sub);
      Tcl_DStringFree(&dStat2Sub);

      for (samp = sampBegin; samp <= sampEnd; samp++)
      {
	/* Write the time (timestamp and seconds) in strings */
	tsStampToText(&pSspec->pTimeStamp[samp], TS_TEXT_MMDDYY, stampText);
	sprintf(text, "%lg", pSspec->pDeltaSec[samp]);

	special = NULL;

	if (pSChan->pData == NULL)
		special = "no_data";
	else if (pSChan->pFlags[samp].eof)
		special = "EOF";
	else if (pSChan->pFlags[samp].missing)
		special = "no_data";
	else if (pSChan->pDataCodeL[samp] == 'I')
		special = "invalid";
	else {}

	if (special != NULL)
	{
	    /* sprintf(datastatL, "%c", pSChan->pDataCodeL[samp]);
	       sprintf(datastatR, "%c", pSChan->pDataCodeR[samp]); */

	    /* This becomes "dataStatus1(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&dStat1Sub, datastatL);

	    /* This becomes "dataStatus2(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&dStat2Sub, datastatR);

	    /* printf("ARR_getRegularData: %s\n", special); */
	    continue;
	}
	else
	{
	    if (type == DBR_TIME_STRING) {
		myType = 2;
		strVal = ((char *)pSChan->pData) +
		    samp * db_strval_dim * pSChan->elCount + sub * db_strval_dim;
	    }
	    else if (type == DBR_TIME_FLOAT) {
		float	*p;
		myType = 1;
		p = ((float *)pSChan->pData) + samp * pSChan->elCount + sub;
		dblVal = (double)(*p);
	    }
	    else if (type == DBR_TIME_SHORT) {
		short	*p;
		myType = 0;
		p = ((short *)pSChan->pData) + samp * pSChan->elCount + sub;
		if (pSChan->isRVAL)
		    lngVal = (long)(*(unsigned short *)p);
		else
		    lngVal = (long)(*p);
	    }
	    else if (type == DBR_TIME_DOUBLE) {
		double	*p;
		myType = 1;
		p = ((double *)pSChan->pData) + samp * pSChan->elCount + sub;
		dblVal = *p;
	    }
	    else if (type == DBR_TIME_LONG) {
		long	*p;
		myType = 0;
		p = ((long *)pSChan->pData) + samp * pSChan->elCount + sub;
		if (pSChan->isRVAL)
		    lngVal = *(unsigned long *)p;
		else
		    lngVal = *p;
	    }
	    else if (type == DBR_TIME_CHAR) {
		char	*p;
		myType = 0;
		p = ((char *)pSChan->pData) + samp * pSChan->elCount + sub;
		lngVal = (long)(*p);
	    }
	    else if (type == DBR_TIME_ENUM) {
		short	*p;
		myType = 0;
		p = ((short *)pSChan->pData) + samp * pSChan->elCount + sub;
		lngVal = (long)(*p);
		if (lngVal < pSChan->grBuf.genmval.no_str)
		    myType = 2, strVal = pSChan->grBuf.genmval.strs[lngVal];
	    }
	    else {}

	    /* Convert data in string format for tcl/tk */
	    if (myType == 0)
	    {
		/* printf("Val = %ld\n", lngVal); */
		sprintf(data, "%ld", lngVal);
	    }
	    else if (myType == 1)
	    {
		/* printf("Val = %g\n", dblVal); */
		sprintf(data, "%lg", dblVal);
	    }
	    else
	    {
		/* strncpy(,,9) with null termination */
		/* Only look at the first 9 chars for brevity */
		int cnt;
                for (cnt=0; cnt<9; cnt++) {
			data[cnt] = strVal[cnt];
			if (strVal[cnt] == '\0')
				break;
		}
		data[9] = '\0';
		/*printf("arr_tcl: %s, strVal = %s\n", data, strVal);*/
	    }

	    sprintf(datastatL, "%c", pSChan->pDataCodeL[samp]);
	    sprintf(datastatR, "%c", pSChan->pDataCodeR[samp]);

/* if (pSChan->pFlags[samp].filled)
   printf("filled  "); */

	    /* stick data into tcl dynamic strings for storage */

	    /* This becomes "value(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&valuesSub, data);

	    /* This becomes "time_sec(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&timesSub, text);

	    /* This becomes "time_ts(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&tStampsSub, stampText);

	    /* This becomes "dataStatus1(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&dStat1Sub, datastatL);

	    /* This becomes "dataStatus2(pvname)" - an arr list */
	    Tcl_DStringAppendElement(&dStat2Sub, datastatR);

	    /*********************************************/
	    /* Determine the maxs & mins of certain data */
	    /*********************************************/

	    /* only if numeric data */
	    if (((myType == 0)||(myType == 1))&&
		 (strcmp(pArrCtx->retrievalCode, "NUMBER+") != 0))
	    {
		/* convert long to dbl */
		if (myType == 0) dblVal = lngVal;

		if (minVal > dblVal)
		    minVal = dblVal;
		if (maxVal < dblVal)
		    maxVal = dblVal;
	    }
	}
      } /* Go on to the next sample */

      /*********************************/
      /* find/convert max & min values */
      /*********************************/
      /* if a "normal" retrieval, max/min of a channel are given by: */
      if (strcmp(pArrCtx->retrievalCode, "NUMBER+") == 0)
      {
	    minVal = pSChan->minDataVal;
	    maxVal = pSChan->maxDataVal;
      }

      /* if non-numeric data, use dummy values as spacers in tcl list */
      else if ((type == DBR_TIME_STRING)||(type == DBR_TIME_ENUM))
      {
	    minVal = 0.;
	    maxVal = 0.;
      }

      /* This string becomes "minimum" - an arr list */
      sprintf(text, "%lg", minVal);

      /* This string becomes "maximum" - an arr list */
      sprintf(data, "%lg", maxVal);

      /* Place temporary list into complete list (as an element). */
      Tcl_DStringAppendElement(&maxs,	data);
      Tcl_DStringAppendElement(&mins,	text);
      Tcl_DStringAppendElement(&values,	Tcl_DStringValue(&valuesSub));
      Tcl_DStringAppendElement(&times,	Tcl_DStringValue(&timesSub));
      Tcl_DStringAppendElement(&tStamps,Tcl_DStringValue(&tStampsSub));
      Tcl_DStringAppendElement(&dStat1,	Tcl_DStringValue(&dStat1Sub));
      Tcl_DStringAppendElement(&dStat2,	Tcl_DStringValue(&dStat2Sub));

    } /* Go on to the next channel */

    /* put all lists together in a list of lists (of lists) */
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&pvNames));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&masterTstamps));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&masterTimes));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&tStamps));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&times));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&values));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&units));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&mins));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&maxs));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&types));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&dStat1));
    Tcl_DStringAppendElement(&pArrCtx->returnValue, Tcl_DStringValue(&dStat2));

    /* Free up memory */
    Tcl_DStringFree(&pvNames);
    Tcl_DStringFree(&masterTstamps);
    Tcl_DStringFree(&masterTimes);
    Tcl_DStringFree(&tStamps);
    Tcl_DStringFree(&times);
    Tcl_DStringFree(&values);
    Tcl_DStringFree(&units);
    Tcl_DStringFree(&mins);
    Tcl_DStringFree(&maxs);
    Tcl_DStringFree(&types);
    Tcl_DStringFree(&dStat1);
    Tcl_DStringFree(&dStat2);

    Tcl_DStringFree(&valuesSub);
    Tcl_DStringFree(&timesSub);
    Tcl_DStringFree(&tStampsSub);
    Tcl_DStringFree(&dStat1Sub);
    Tcl_DStringFree(&dStat2Sub);

    /* return value from this routine */
    return TCL_OK;
}


/********************************************************
* NAME	posBOF - position file at the beginning
*
*********************************************************/
#ifdef __STDC__
static long
posBOF(FILE *pStream)
#else
static long
posBOF(pStream)
  FILE	*pStream;
#endif /*__STDC__*/

{
    if (fseek(pStream, 0L, 0) < 0) {
	perror("posBOF: fseek error");
	return -1;
    }
    return ftell(pStream);
}

/***********************************************************
* NAME	posEOF - position file at the end
*
************************************************************/
#ifdef __STDC__
static long
posEOF(FILE *pStream)
#else
static long
posEOF(pStream)
  FILE	*pStream;
#endif /*__STDC__*/

{
    if (fseek(pStream, 0L, 2) < 0) {
	perror("posEOF: fseek error");
	return -1;
    }
    return ftell(pStream);
}


/***********************************************************
* NAME	posKeyFwd - search forward for a string
*
* RETURNS
*	position of key (its first character), or
*	-1 if search fails
*
************************************************************/
#ifdef __STDC__
static long
posKeyFwd(char *key, FILE *pStream)
#else
static long
posKeyFwd(key, pStream)
  char	*key;
  FILE	*pStream;
#endif /*__STDC__*/

{
    int		i, lastChar;
    int		c;
    long	pos;

    lastChar = strlen(key) - 1;
    i = 0;

    if ((pos = ftell(pStream)) < 0) {
	if (!feof(pStream))
	    perror("posKeyFwd: ftell error");
	return -1;
    }

    while (1) {
	if ((c = fgetc(pStream)) == EOF)
	    return -1;
	if ((char)c == key[i]) {
	    if (i == lastChar)
		break;
	    i++;
	}
	else {
	    if (i == 0) {
		if ((pos = ftell(pStream)) < 0) {
		    if (!feof(pStream))
			perror("posKeyFwd: ftell error");
		    return -1;
		}
	    }
	    else if (i == 1) {
		pos++;
		if ((char)c != key[0])
		    i = 0;
	    }
	    else {
		pos++;
		if (fseek(pStream, pos, 0) != 0) {
		    if (!feof(pStream))
			perror("posKeyFwd: fseek error");
		    return -1;
		}
		i = 0;
	    }
	}
    }
    if (fseek(pStream, pos, 0) != 0) {
	if (!feof(pStream))
	    perror("posKeyFwd: fseek error");
	return -1;
    }
    return pos;
}


/*****************************************************
* NAME	posKeyRev - search backward for a string
*
* RETURNS
*	position of key (its first character), or
*	-1 if search fails
*
******************************************************/
#ifdef __STDC__
static long
posKeyRev(
  char	*key,
  FILE	*pStream,
  long	lastPos)	/* I position to start search; -1 for EOF */
#else
static long
posKeyRev(key, pStream, lastPos)
  char	*key;
  FILE	*pStream;
  long	lastPos;	/* I position to start search; -1 for EOF */
#endif /*__STDC__*/

{
    int		i, lastChar;
    int		c;
    long	pos;

    lastChar = i = strlen(key) - 1;

    if (lastPos >= 0) {
	if (fseek(pStream, lastPos, 0) != 0) {
	    if (!feof(pStream))
		perror("posKeyRev: fseek error");
	    return -1;
	}
    }
    else {
	if (fseek(pStream, 0L, 2) != 0) {
	    if (!feof(pStream))
		perror("posKeyRev: fseek error");
	    return -1;
	}
    }
    while (1) {
	if (ftell(pStream) <= 0)
	    return -1;
	if (fseek(pStream, -1L, 1) != 0) {
	    if (!feof(pStream))
		perror("posKeyRev: fseek error");
	    return -1;
	}
	if ((c = fgetc(pStream)) == EOF)
	    return -1;
	if (fseek(pStream, -1L, 1) != 0) {
	    if (!feof(pStream))
		perror("posKeyRev: fseek error");
	    return -1;
	}
	if ((char)c == key[i]) {
	    if (i == 0)
		break;
	    i--;
	}
	else
	    i = lastChar;
    }
    if ((pos = ftell(pStream)) < 0)
	perror("posKeyRev: ftell error");
    return pos;
}

/*****************************************************
* NAME	posPos - set file to desired position
*
******************************************************/
#ifdef __STDC__
static long
posPos(
  FILE	*pStream,
  long	pos)
#else
static long
posPos(pStream, pos)
  FILE	*pStream;
  long	pos;
#endif /*__STDC__*/

{
    if (fseek(pStream, pos, 0) < 0) {
	perror("posPos: fseek error");
	return -1;
    }
    return ftell(pStream);
}


/************************************************************************
* NAME	sydSSFFuncSeekStamp - find a `SAMPLE: at' line
*
* DESCRIPTION
*	Find a `SAMPLE: at' line and get the time stamp from it.
*
*	If the search is forward, the file is left positioned following
*	the stamp.  If backward, the position is at the beginning of the
*	line.
*
* RETURNS
*	position of the line
*
**************************************************************************/
#ifdef __STDC__
static long
sydSSFFuncSeekStamp(
  FILE		*ssFile,
  TS_STAMP	*pStamp,
  int		revFlag,	/* I 1 says seek in reverse direction */
  long		pos)		/* I position to start search if reverse;
				   -1 says to start at EOF */
#else
static long
sydSSFFuncSeekStamp(ssFile, pStamp, revFlag, pos)
  FILE		*ssFile;
  TS_STAMP	*pStamp;
  int		revFlag;	/* I 1 says seek in reverse direction */
  long		pos;		/* I position to start search if reverse;
				   -1 says to start at EOF */
#endif /*__STDC__*/
{
    long	bufStat;
    char	record[120], recDelim;
    char	*pRecord, *pField, delim;
    long	stat;

    if (revFlag)
	pos = posKeyRev("SAMPLE: at", ssFile, pos);
    else 
	pos = posKeyFwd("SAMPLE: at", ssFile);
    if (pos < 0)
	return pos;
    assert(sizeof(record) == 120);
    if (fgets(record, 120, ssFile) == NULL)
	return -1;

    pRecord = record;
    recDelim = record[strlen(record)-1];
    if (nextNonSpaceField(&pRecord, &pField, &delim) <= 1)
	assertAlways(0);
    if (strcmp(pField, "SAMPLE:") != 0)
	assertAlways(0);
    if (nextNonSpaceField(&pRecord, &pField, &delim) <= 1)
	assertAlways(0);
    if (strcmp(pField, "at") != 0)
	assertAlways(0);
    stat = tsTextToStamp(pStamp, &pRecord);
    assert(stat == S_ts_OK);
    if (revFlag) {
	if (posPos(ssFile, pos) < 0)
	    assertAlways(0);
    }
    return pos;
}
